import math
from typing import Dict

import numpy as np

from autoconf.dictable import Dictable

"""
The `Gaussian` class in this module is the model components that is fitted to data using a non-linear search. The
inputs of its __init__ constructor are the parameters which can be fitted for.

The log_likelihood_function in the Analysis class receives an instance of this classes where the values of its
parameters have been set up according to the non-linear search. Because instances of the classes are used, this means
their methods (e.g. model_data_1d_via_xvalues_from) can be used in the log likelihood function.
"""


class Gaussian(Dictable):
    def __init__(
        self,
        centre: float = 0.0,  # <- PyAutoFit recognises these constructor arguments
        normalization: float = 0.1,  # <- are the Gaussian`s model parameters.
        sigma: float = 0.01,
    ):
        """
        Represents a 1D `Gaussian` profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a non-linear search.

        Parameters
        ----------
        centre
            The x coordinate of the profile centre.
        normalization
            Overall normalization normalisation of the `Gaussian` profile.
        sigma
            The sigma value controlling the size of the Gaussian.
        """
        self.centre = centre
        self.normalization = normalization
        self.sigma = sigma

    def model_data_1d_via_xvalues_from(self, xvalues: np.ndarray) -> np.ndarray:
        """
        Calculate the normalization of the profile on a 1D grid of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Gaussian, using its centre.

        Parameters
        ----------
        xvalues
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = xvalues - self.centre

        return np.multiply(
            np.divide(self.normalization, self.sigma * np.sqrt(2.0 * np.pi)),
            np.exp(-0.5 * np.square(np.divide(transformed_xvalues, self.sigma))),
        )

    def __call__(self, xvalues: np.ndarray) -> np.ndarray:
        """
        For certain graphical models, the `__call__` function is overwritten for producing the model-fit.
        We include this here so these examples work, but it should not be important for most PyAutoFit users.

        Parameters
        ----------
        xvalues
            The x coordinates in the original reference frame of the grid.
        """
        return self.model_data_1d_via_xvalues_from(xvalues=xvalues)

    def dict(self) -> Dict:
        """
        Returns the `Gaussian` as a dictionary which can be straight forwardly written to a `.json` file via the
        code:

        with open(model_file, "w+") as f:
            json.dump(gaussian.dict(), f, indent=4)

        Returns
        -------
        The `Gaussian` type and model parameters as a dictionary.
        """
        return super().dict()

    def inverse(self, y):
        """
        For graphical models, the inverse of the Gaussian is used to test certain aspects of the calculation.
        """

        a = self.normalization / (y * self.sigma * math.sqrt(2 * math.pi))

        b = 2 * math.log(a)

        return self.centre + self.sigma * math.sqrt(b)


class Exponential(Dictable):
    def __init__(
        self,
        centre: float = 0.0,  # <- PyAutoFit recognises these constructor arguments are the model
        normalization: float = 0.1,  # <- parameters of the Gaussian.
        rate: float = 0.01,
    ):
        """
        Represents a 1D Exponential profile, which may be treated as a model-component of PyAutoFit the
        parameters of which are fitted for by a `NonLinearSearch`.

        Parameters
        ----------
        centre
            The x coordinate of the profile centre.
        normalization
            Overall normalization normalisation of the `Gaussian` profile.
        rate
            The decay rate controlling has fast the Exponential declines.
        """
        self.centre = centre
        self.normalization = normalization
        self.rate = rate

    def model_data_1d_via_xvalues_from(self, xvalues: np.ndarray) -> np.ndarray:
        """
        Calculate the 1D Gaussian profile on a 1D grid of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values
            The x coordinates in the original reference frame of the grid.
        """
        transformed_xvalues = np.subtract(xvalues, self.centre)
        return self.normalization * np.multiply(
            self.rate, np.exp(-1.0 * self.rate * abs(transformed_xvalues))
        )

    def __call__(self, xvalues: np.ndarray) -> np.ndarray:
        """
        Calculate the 1D Gaussian profile on a 1D grid of Cartesian x coordinates.

        The input xvalues are translated to a coordinate system centred on the Exponential, using its centre.

        Parameters
        ----------
        values
            The x coordinates in the original reference frame of the grid.
        """
        return self.model_data_1d_via_xvalues_from(xvalues=xvalues)

    def dict(self) -> Dict:
        """
        Returns the `Gaussian` as a dictionary which can be straight forwardly written to a `.json` file via the
        code:

        with open(model_file, "w+") as f:
            json.dump(gaussian.dict(), f, indent=4)

        Returns
        -------
        The `Gaussian` type and model parameters as a dictionary.

        """
        return super().dict()



from autofit.plot.samples_plotters import MCMCPlotter


class ZeusPlotter(MCMCPlotter):

    def corner(self, **kwargs):

        import zeus

        try:
            zeus.cornerplot(
                samples=self.samples.results_internal.get_chain(flat=True),
                weight_list=self.samples.weight_list,
                labels=self.model.parameter_labels_with_superscripts_latex,
                **kwargs
            )
        except TypeError:
            pass

        self.output.to_figure(structure=None, auto_filename="corner")

    def trajectories(self, **kwargs):

        self._plot_trajectories(
            samples=self.samples.results_internal.get_chain(),
            log_posterior_list=self.samples.results_internal.get_log_prob(),
            **kwargs
        )

    def likelihood_series(self, **kwargs):

        self._plot_likelihood_series(
            log_posterior_list = self.samples.results_internal.get_log_prob()
        )

    def time_series(self, **kwargs):

        self._plot_time_series(
            samples=self.samples.results_internal.get_chain()
        )


import logging
import os
import pickle
from pathlib import Path
from typing import Optional, Union

from .. import model as m
from ..sqlalchemy_ import sa
from ...mapper.model_object import Identifier

logger = logging.getLogger(__name__)


def _parent_identifier(directory: str) -> Optional[str]:
    """
    Read the parent identifier for a fit in a directory.

    Defaults to None if no .parent_identifier file is found.
    """
    try:
        with open(f"{directory}/.parent_identifier") as f:
            return f.read()
    except FileNotFoundError:
        return None


class Scraper:
    def __init__(self, directory: Union[Path, str], session: sa.orm.Session):
        """
        Facilitates scraping of data output into a directory
        into the database.

        Parameters
        ----------
        directory
            A directory in which data has been stored
        session
            A database session
        """
        self.directory = directory
        self.session = session

    def scrape(self):
        """
        Recursively scrape fits from the directory and
        add them to the session
        """
        for fit in self._fits():
            self.session.add(fit)
        for grid_search in self._grid_searches():
            self.session.add(grid_search)

    def _fits(self):
        """
        Scrape data output into a directory tree so it can be added to the
        aggregator database.

        Returns
        -------
        Generator yielding Fit database objects
        """
        logger.info(f"Scraping directory {self.directory}")
        from autofit.aggregator.aggregator import Aggregator as ClassicAggregator

        aggregator = ClassicAggregator(self.directory)
        logger.info(f"{len(aggregator)} searches found")
        for item in aggregator:
            is_complete = os.path.exists(f"{item.directory}/.completed")

            parent_identifier = _parent_identifier(directory=item.directory)

            model = item.model
            samples = item.samples

            identifier = _make_identifier(item)

            logger.info(
                f"Creating fit for: "
                f"{item.search.paths.path_prefix} "
                f"{item.search.unique_tag} "
                f"{item.search.name} "
                f"{identifier} "
            )

            try:
                instance = samples.max_log_likelihood()
            except (AttributeError, NotImplementedError):
                instance = None

            try:
                fit = self._retrieve_model_fit(item)
                logger.warning(f"Fit already existed with identifier {identifier}")
            except sa.orm.exc.NoResultFound:
                try:
                    log_likelihood = samples.max_log_likelihood_sample.log_likelihood
                except AttributeError:
                    log_likelihood = None
                fit = m.Fit(
                    id=identifier,
                    name=item.search.name,
                    unique_tag=item.search.unique_tag,
                    model=model,
                    instance=instance,
                    is_complete=is_complete,
                    info=item.info,
                    max_log_likelihood=log_likelihood,
                    parent_id=parent_identifier,
                )

            pickle_path = Path(item.pickle_path)
            _add_pickles(fit, pickle_path)
            for i, child_analysis in enumerate(item.child_analyses):
                child_fit = m.Fit(
                    id=f"{identifier}_{i}",
                )
                _add_pickles(child_fit, child_analysis.pickle_path)
                fit.children.append(child_fit)

            yield fit

    def _grid_searches(self):
        """
        Retrieve grid searches recursively from an output directory by
        searching for the .is_grid_search file.

        Should be called after adding Fits as it relies on querying fits

        Yields
        ------
        Fit objects representing grid searches with child fits associated
        """
        from autofit.aggregator.aggregator import Aggregator as ClassicAggregator

        for root, _, filenames in os.walk(self.directory):
            if ".is_grid_search" in filenames:
                path = Path(root)

                is_complete = (path / ".completed").exists()

                with open(path / ".is_grid_search") as f:
                    unique_tag = f.read()

                grid_search = m.Fit(
                    id=path.name,
                    unique_tag=unique_tag,
                    is_grid_search=True,
                    parent_id=_parent_identifier(root),
                    is_complete=is_complete,
                )

                pickle_path = path / "pickles"
                _add_pickles(grid_search, pickle_path)

                aggregator = ClassicAggregator(root)
                for item in aggregator:
                    fit = self._retrieve_model_fit(item)
                    grid_search.children.append(fit)
                yield grid_search

    def _retrieve_model_fit(self, item) -> m.Fit:
        """
        Retrieve a Fit, if one exists, corresponding to a given SearchOutput

        Parameters
        ----------
        item
            A SearchOutput from the classic Aggregator

        Returns
        -------
        A fit with the corresponding identifier

        Raises
        ------
        NoResultFound
            If no fit is found with the identifier
        """
        return (
            self.session.query(m.Fit).filter(m.Fit.id == _make_identifier(item)).one()
        )


def _make_identifier(item) -> str:
    """
    Create a unique identifier for a SearchOutput.

    This accounts for the Search, Model and unique_tag

    Parameters
    ----------
    item
        An output from the classic aggregator

    Returns
    -------
    A unique identifier that is sensitive to changes that affect
    the search
    """
    search = item.search
    model = item.model
    return str(Identifier([search, model, search.unique_tag]))


def _add_pickles(fit: m.Fit, pickle_path: Path):
    """
    Load pickles from the path and add them to the database.

    Parameters
    ----------
    fit
        A fit to which the pickles belong
    pickle_path
        The path in which the pickles are stored
    """
    try:
        filenames = os.listdir(pickle_path)
    except FileNotFoundError as e:
        logger.exception(e)
        filenames = []

    for filename in filenames:
        try:
            with open(pickle_path / filename, "r+b") as f:
                fit[filename.split(".")[0]] = pickle.load(f)
        except (pickle.UnpicklingError, ModuleNotFoundError) as e:
            if filename == "dynesty.pickle":
                continue

            raise pickle.UnpicklingError(
                f"Failed to unpickle: {pickle_path} {filename}"
            ) from e



import logging
from abc import ABC
import os

from autoconf import conf

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.result import Result

logger = logging.getLogger(
    __name__
)


class Analysis(ABC):
    """
    Protocol for an analysis. Defines methods that can or
    must be implemented to define a class that compute the
    likelihood that some instance fits some data.
    """

    def with_model(self, model):
        """
        Associate an explicit model with this analysis. Instances of the model
        will be used to compute log likelihood in place of the model passed
        from the search.

        Parameters
        ----------
        model
            A model to associate with this analysis

        Returns
        -------
        An analysis for that model
        """
        from .model_analysis import ModelAnalysis
        return ModelAnalysis(
            analysis=self,
            model=model
        )

    def should_visualize(self, paths: AbstractPaths, during_analysis : bool = True) -> bool:
        """
        Whether a visualize method should be called perform visualization, which depends on the following:

        1) If a model-fit has already completed, the default behaviour is for visualization to be bypassed in order
        to make model-fits run faster.

        2) If a model-fit has completed, but it is the final visualization output where `during_analysis` is False,
        it should be performed.

        3) Visualization can be forced to run via the `force_visualization_overwrite`, for example if a user
        wants to plot additional images that were not output on the original run.

        4) If PyAutoFit test mode is on visualization is disabled, irrespective of the `force_visualization_overwite`
        config input.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization and the pickled objects used by the aggregator output by this function.


        Returns
        -------
        A bool determining whether visualization should be performed or not.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            return False

        if conf.instance["general"]["output"]["force_visualize_overwrite"]:
            return True

        if not during_analysis:
            return True

        return not paths.is_complete

    def log_likelihood_function(self, instance):
        raise NotImplementedError()

    def visualize_before_fit(self, paths: AbstractPaths, model: AbstractPriorModel):
        pass

    def visualize(self, paths: AbstractPaths, instance, during_analysis):
        pass

    def visualize_before_fit_combined(self, analyses, paths: AbstractPaths, model: AbstractPriorModel):
        pass

    def visualize_combined(self, analyses, paths: AbstractPaths, instance, during_analysis):
        pass

    def save_attributes_for_aggregator(self, paths: AbstractPaths):
        pass

    def save_results_for_aggregator(self, paths: AbstractPaths, result:Result):
        pass

    def modify_before_fit(self, paths: AbstractPaths, model: AbstractPriorModel):
        """
        Overwrite this method to modify the attributes of the `Analysis` class before the non-linear search begins.

        An example use-case is using properties of the model to alter the `Analysis` class in ways that can speed up
        the fitting performed in the `log_likelihood_function`.
        """
        return self

    def modify_model(self, model):
        return model

    def modify_after_fit(self, paths: AbstractPaths, model: AbstractPriorModel, result: Result):
        """
        Overwrite this method to modify the attributes of the `Analysis` class before the non-linear search begins.

        An example use-case is using properties of the model to alter the `Analysis` class in ways that can speed up
        the fitting performed in the `log_likelihood_function`.
        """
        return self

    def make_result(self, samples, model, sigma=1.0, use_errors=True, use_widths=False):
        return Result(
            samples=samples,
            model=model,
            sigma=sigma,
            use_errors=use_errors,
            use_widths=use_widths,
        )

    def profile_log_likelihood_function(self, paths: AbstractPaths, instance):
        """
        Overwrite this function for profiling of the log likelihood function to be performed every update of a 
        non-linear search.
        
        This behaves analogously to overwriting the `visualize` function of the `Analysis` class, whereby the user 
        fills in the project-specific behaviour of the profiling.
        
        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likliehood instance of the model so far in the non-linear search.
        """
        pass

    def __add__(
            self,
            other: "Analysis"
    ):
        """
        Analyses can be added together. The resultant
        log likelihood function returns the sum of the
        underlying log likelihood functions.

        Parameters
        ----------
        other
            Another analysis class

        Returns
        -------
        A class that computes log likelihood based on both analyses
        """
        from .combined import CombinedAnalysis
        if isinstance(
                other,
                CombinedAnalysis
        ):
            return other + self
        return CombinedAnalysis(
            self, other
        )

    def __radd__(self, other):
        """
        Allows analysis to be used in sum
        """
        if other == 0:
            return self
        return self + other



from collections import Counter, defaultdict
from functools import reduce
from itertools import count
from typing import Tuple, Dict, Collection, List, Type

import numpy as np

from autoconf import cached_property
from autofit.graphical.factor_graphs.abstract import FactorValue, AbstractNode
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.jacobians import VectorJacobianProduct
from autofit.graphical.utils import rescale_to_artists
from autofit.mapper.variable import Variable


class FactorGraph(AbstractNode):
    def __init__(
            self,
            factors: Collection[Factor],
    ):
        """
        A graph relating factors

        Parameters
        ----------
        factors
            Nodes wrapping individual factors in a model
        """
        self._name = "(%s)" % "*".join(f.name for f in factors)

        self._factors = tuple(factors)

        self._factor_all_variables = {f: f.all_variables for f in self._factors}

        self._call_sequence = self._get_call_sequence()

        self._validate()

        _kwargs = {variable.name: variable for variable in self.variables}

        super().__init__(**_kwargs)

    def copy(self):
        return type(self)(self.factors)

    def related_factors(self, variable: Variable, excluded_factor=None) -> List[Factor]:
        """
        A list of factors which contain the variable.

        Parameters
        ----------
        excluded_factor
            A factor that should be excluded from the list
        variable
            A variable in the graph which will be related to one
            or more factors

        Returns
        -------
        The factors associated with the variable
        """
        return sorted(
            {
                factor
                for factor in self._factors
                if variable in factor.variables
                   and (excluded_factor is None or factor != excluded_factor)
            }
        )

    def _factors_with_type(self, factor_type: Type[Factor]) -> List[Factor]:
        """
        Find all factors with a given type
        """
        return [factor for factor in self._factors if isinstance(factor, factor_type)]

    def factors_by_type(self) -> Dict[Type[Factor], List[Factor]]:
        """
        A dictionary mapping types of factor to all factors of that type
        """
        factors_by_type = defaultdict(list)
        for factor in self._factors:
            factors_by_type[type(factor)].append(factor)
        return factors_by_type

    @property
    def info(self) -> str:
        """
        Describes the graph. Output in graph.info
        """
        string = ""
        for factor_type, factors in self.factors_by_type().items():
            factor_info = "\n\n".join(factor.info for factor in factors)
            string = f"{string}{factor_type.__name__}s\n\n{factor_info}\n\n"
        return string

    def make_results_text(self, model_approx) -> str:
        """
        Generate text describing the graph w.r.t. a given model approximation
        """
        results_text = "\n\n".join(
            factor.make_results_text(model_approx) for factor in self._factors
        )
        return f"{self.name}\n\n{results_text}"

    @property
    def name(self):
        return self._name

    def _validate(self):
        """
        Raises
        ------
        If there is an inconsistency with this graph
        """
        det_var_counts = ", ".join(
            map(
                str,
                [
                    v
                    for v, c in Counter(
                    v for f in self.factors for v in f.deterministic_variables
                ).items()
                    if c > 1
                ],
            )
        )
        if det_var_counts:
            raise ValueError(
                "Improper FactorGraph, "
                f"Deterministic variables {det_var_counts} appear in "
                "multiple factors"
            )

    @cached_property
    def all_variables(self):
        return reduce(
            frozenset.union, (factor.all_variables for factor in self.factors)
        )

    @cached_property
    def deterministic_variables(self):
        return reduce(
            frozenset.union, (factor.deterministic_variables for factor in self.factors)
        )

    @cached_property
    def variables(self):
        return self.all_variables - self.deterministic_variables

    def _get_call_sequence(self) -> List[List[Factor]]:
        """
        Compute the order in which the factors must be evaluated. This is done by checking whether
        all variables required to call a factor are present in the set of variables encapsulated
        by all factors, not including deterministic variables.

        Deterministic variables must be computed before the dependent factors can be computed.
        """
        call_sets = defaultdict(list)
        for factor in self.factors:
            missing_vars = frozenset(factor.variables.difference(self.variables))
            call_sets[missing_vars].append(factor)

        call_sequence = []
        while call_sets:
            # the factors that can be evaluated have no missing variables
            factors = call_sets.pop(frozenset(()))
            # if there's a KeyError then the FactorGraph is improper
            calls = []
            new_variables = set()
            for factor in factors:
                det_vars = factor.deterministic_variables
                calls.append(factor)
                # TODO: this might cause problems
                # TODO: if det_vars appear more than once
                new_variables.update(det_vars)

            call_sequence.append(calls)

            # update to include newly calculated factors
            for missing in list(call_sets.keys()):
                if missing.intersection(new_variables):
                    factors = call_sets.pop(missing)
                    call_sets[missing.difference(new_variables)].extend(factors)

        return call_sequence

    def _call_args(self, *args):
        return self(dict(zip(self.args, args)))

    def __call__(
            self,
            variable_dict: Dict[Variable, np.ndarray],
    ) -> FactorValue:
        """
        Call each function in the graph in the correct order, adding the logarithmic results.

        Deterministic values computed in initial factor calls are added to a dictionary and
        passed to subsequent factor calls.

        Parameters
        ----------
        variable_dict
            Positional arguments
        axis
            Keyword arguments

        Returns
        -------
        Object comprising the log value of the computation and a dictionary containing
        the values of deterministic variables.
        """

        # generate set of factors to call, these are indexed by the
        # missing deterministic variables that need to be calculated
        log_value = 0.0
        det_values = {}
        variables = {**self.fixed_values, **variable_dict}

        missing = set(v.name for v in self.variables).difference(
            v.name for v in variables
        )
        if missing:
            n_miss = len(missing)
            missing_str = ", ".join(missing)
            raise ValueError(
                f"{self} missing {n_miss} arguments: {missing_str}"
                f"factor graph call signature: {self.call_signature}"
            )

        for calls in self._call_sequence:
            # TODO parallelise this part?
            for factor in calls:
                ret = factor(variables)
                log_value += np.sum(ret)
                det_values.update(ret.deterministic_values)
                variables.update(ret.deterministic_values)

        return FactorValue(log_value, det_values)

    def func_jacobian(
            self, variable_dict: Dict[Variable, np.ndarray], **kwargs
    ) -> FactorValue:

        # generate set of factors to call, these are indexed by the
        # missing deterministic variables that need to be calculated
        log_value = 0.0
        det_values = {}
        factor_jacs = {}
        variables = {**self.fixed_values, **variable_dict}

        missing = set(v.name for v in self.variables).difference(
            v.name for v in variables
        )
        if missing:
            n_miss = len(missing)
            missing_str = ", ".join(missing)
            raise ValueError(
                f"{self} missing {n_miss} arguments: {missing_str}"
                f"factor graph call signature: {self.call_signature}"
            )

        for calls in self._call_sequence:
            # TODO parallelise this part?
            for factor in calls:
                ret, factor_jacs[factor] = factor.func_jacobian(variables, **kwargs)
                log_value += ret

                det_values.update(ret.deterministic_values)
                variables.update(ret.deterministic_values)

        jac_out = (FactorValue,) + tuple(det_values)
        jac_variables = tuple(variables)

        def graph_vjp(args):
            args = args if isinstance(args, tuple) else (args,)
            grads = {}
            grads.update(zip(jac_out, args))
            for calls in self._call_sequence[::-1]:
                for factor in calls:
                    factor_grad = factor_jacs[factor](grads)
                    for v, v_grad in factor_grad.items():
                        grads[v] = grads.get(v, 0) + v_grad

            return tuple(grads[v] for v in jac_variables)

        fval = FactorValue(log_value, det_values)
        graph_vjp = VectorJacobianProduct(
            jac_out,
            graph_vjp,
            *jac_variables,
            out_shapes=fval.to_dict().shapes,
        )

        return fval, graph_vjp

    def __mul__(self, other: AbstractNode) -> "FactorGraph":
        """
        Combine this object with another factor node or graph, creating
        a new graph that comprises all of the factors of the two objects.
        """
        factors = self.factors

        if isinstance(other, FactorGraph):
            factors += other.factors
        elif isinstance(other, Factor):
            factors += (other,)
        else:
            raise TypeError(
                f"type of passed element {(type(other))} "
                "does not match required types, (`FactorGraph`, `FactorNode`)"
            )

        return type(self)(factors)

    def __repr__(self) -> str:
        factors_str = " * ".join(map(repr, self.factors))
        return f"({factors_str})"

    @property
    def factors(self) -> Tuple[Factor, ...]:
        return self._factors

    @property
    def factor_all_variables(self) -> Dict[Factor, List[Variable]]:
        return self._factor_all_variables

    @property
    def graph(self):
        try:
            import networkx as nx
        except ImportError as e:
            raise ImportError("networkx required for graph") from e

        G = nx.Graph()
        G.add_nodes_from(self.factors, bipartite="factor")
        G.add_nodes_from(self.all_variables, bipartite="variable")
        G.add_edges_from(
            (f, v) for f, vs in self.factor_all_variables.items() for v in vs
        )

        return G

    def draw_graph(
            self,
            pos=None,
            ax=None,
            size=20,
            color="k",
            fill="w",
            factor_shape="s",
            variable_shape="o",
            factor_labels=None,
            variable_labels=None,
            factor_kws=None,
            variable_kws=None,
            edge_kws=None,
            factors=None,
            draw_labels=False,
            label_kws=None,
            **kwargs,
    ):
        try:
            import matplotlib.pyplot as plt
            import networkx as nx
        except ImportError as e:
            raise ImportError(
                "Matplotlib and networkx required for draw_graph()"
            ) from e
        except RuntimeError as e:
            print("Matplotlib unable to open display")
            raise e

        if ax is None:
            ax = plt.gca()

        G = self.graph
        if pos is None:
            pos = bipartite_layout(factors or self.factors)

        kwargs.setdefault("ms", size)
        kwargs.setdefault("c", color)
        kwargs.setdefault("mec", color)
        kwargs.setdefault("mfc", fill)
        kwargs.setdefault("ls", "")

        factor_kws = factor_kws or {}
        factor_kws.setdefault("marker", factor_shape)

        variable_kws = variable_kws or {}
        variable_kws.setdefault("marker", variable_shape)

        # draw factors
        xy = np.array([pos[f] for f in self.factors]).T
        fs = ax.plot(*xy, **{**kwargs, **factor_kws})
        # draw variables
        xy = np.array([pos[f] for f in self.all_variables]).T
        vs = ax.plot(*xy, **{**kwargs, **variable_kws})
        # draw edges
        edges = nx.draw_networkx_edges(G, pos, **(edge_kws or {}))

        # remove ticks from axes
        ax.tick_params(
            axis="both",
            which="both",
            bottom=False,
            left=False,
            labelbottom=False,
            labelleft=False,
        )
        if draw_labels:
            self.draw_graph_labels(
                pos,
                ax=ax,
                factor_labels=factor_labels,
                variable_labels=variable_labels,
                **(label_kws or {})
            )
        return pos, fs, vs, edges

    def draw_graph_labels(
            self,
            pos,
            factor_labels=None,
            variable_labels=None,
            shift=0.1,
            f_shift=None,
            v_shift=None,
            f_horizontalalignment="right",
            v_horizontalalignment="left",
            f_kws=None,
            v_kws=None,
            graph=None,
            ax=None,
            rescale=True,
            **kwargs,
    ):
        try:
            import matplotlib.pyplot as plt
            import networkx as nx
        except ImportError as e:
            raise ImportError(
                "Matplotlib and networkx required for draw_graph()"
            ) from e
        ax = ax or plt.gca()
        graph = graph or self.graph
        factor_labels = factor_labels or {f: f.name for f in self.factors}
        variable_labels = variable_labels or {v: v.name for v in self.all_variables}
        f_kws = f_kws or {"horizontalalignment": f_horizontalalignment}
        v_kws = v_kws or {"horizontalalignment": v_horizontalalignment}

        f_shift = f_shift or shift
        f_pos = {f: (x - f_shift, y) for f, (x, y) in pos.items()}
        v_shift = v_shift or shift
        v_pos = {f: (x + v_shift, y) for f, (x, y) in pos.items()}

        text = {
            **nx.draw_networkx_labels(
                graph, f_pos, labels=factor_labels, ax=ax, **f_kws, **kwargs
            ),
            **nx.draw_networkx_labels(
                graph, v_pos, labels=variable_labels, ax=ax, **v_kws, **kwargs
            ),
        }
        if rescale:
            rescale_to_artists(text.values(), ax=ax)

        return text


def bipartite_layout(factors):
    n_factors = len(factors)
    n_variables = len(set().union(*(f.variables for f in factors)))
    n = max(n_factors, n_variables)
    factor_count = count()
    variable_count = count()

    pos = {}
    for factor in factors:
        pos[factor] = 0, next(factor_count) * n / n_factors
        for v in factor.variables:
            if v not in pos:
                pos[v] = 1, next(variable_count) * n / n_variables

    return pos



from typing import List, Union

from autofit.database import Prior
from autofit.non_linear.result import Result, Placeholder
from .job import JobResult
from .result import GridSearchResult


class ResultBuilder:
    def __init__(
            self,
            lists: List[List[float]],
            grid_priors: List[Prior]
    ):
        """
        Builds GridSearchResults including all results so far computed
        and Placeholders where no result has yet been computed.

        Parameters
        ----------
        lists
            A list of lists of unit hypercube vectors describing the
            values that change during the grid search
        grid_priors
            Priors which are varied throughout the grid search
        """
        self.lists = lists
        self.grid_priors = grid_priors
        self._job_result_dict = {}

    def __call__(self) -> GridSearchResult:
        """
        Generate a GridSearchResult with all results so far and placeholders
        where no result has been returned yet
        """
        return GridSearchResult(
            self.results,
            self.lists,
            self.grid_priors
        )

    @property
    def results(self) -> List[Union[Result, Placeholder]]:
        """
        A list of results that have been returned with placeholders where no
        result has been returned in the grid-search order.
        """
        results = []
        for number in range(len(self.lists)):
            try:
                job_result = self._job_result_dict[
                    number
                ]
                result = job_result.result
                results.append(
                    Result(
                        samples=job_result.result.samples,
                        model=job_result.result.model,
                        sigma=result.sigma,
                        use_errors=result.use_errors,
                        use_widths=result.use_widths,
                    )
                )
            except KeyError:
                results.append(
                    Placeholder()
                )
        return results

    def add(self, job_result: JobResult):
        """
        Add a new result for a completed job.

        The number associated with the job result is used for
        tracking and ordering.

        Parameters
        ----------
        job_result
            A result of an optimisation for a single point in
            the grid.
        """
        self._job_result_dict[
            job_result.number
        ] = job_result



import abc
import inspect
from typing import List, Tuple, Any, Iterable, Union, ItemsView, Type

import numpy as np

from autoconf.class_path import get_class, get_class_path
from ..sqlalchemy_ import sa, declarative

Base = declarative.declarative_base()

_schema_version = 1


class Object(Base):
    __tablename__ = "object"

    type = sa.Column(
        sa.String
    )

    id = sa.Column(
        sa.Integer,
        primary_key=True,
    )

    parent_id = sa.Column(
        sa.Integer,
        sa.ForeignKey(
            "object.id"
        )
    )
    parent = sa.orm.relationship(
        "Object",
        uselist=False,
        remote_side=[id]
    )

    samples_for_id = sa.Column(
        sa.String,
        sa.ForeignKey(
            "fit.id"
        )
    )
    samples_for = sa.orm.relationship(
        "Fit",
        uselist=False,
        foreign_keys=[samples_for_id]
    )

    children: List["Object"] = sa.orm.relationship(
        "Object",
        uselist=True,
    )

    def __len__(self):
        return len(self.children)

    name = sa.Column(sa.String)

    __mapper_args__ = {
        'polymorphic_identity': 'object',
        'polymorphic_on': type
    }

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name}>"

    # noinspection PyProtectedMember
    @classmethod
    def from_object(
            cls,
            source,
            name=""
    ):
        """
        Create a database object for an object in a model.

        The specific database class used depends on the type of
        the object.

        If __getstate__ is defined for any instance then that
        dictionary is used in place of the __dict__ when
        serialising.

        Parameters
        ----------
        source
            A model
        name
            The name of the object wrt its parent

        Returns
        -------
        An instance of a concrete child of this class
        """

        from autofit.mapper.prior_model.prior_model import Model
        from autofit.mapper.prior.abstract import Prior
        from autofit.mapper.prior_model.collection import Collection

        if source is None or isinstance(
                source,
                (
                        np.ndarray,
                        np.broadcast,
                        abc.ABCMeta,
                        np.ufunc,
                )
        ):
            from .instance import NoneInstance
            instance = NoneInstance()
        elif isinstance(source, Model):
            from .prior import Model
            instance = Model._from_object(
                source
            )
        elif isinstance(source, Prior):
            from .prior import Prior
            instance = Prior._from_object(
                source
            )
        elif isinstance(source, (float, int)):
            from .instance import Value
            instance = Value._from_object(
                source
            )
        elif isinstance(source, (tuple, list)):
            from .instance import Collection
            instance = Collection._from_object(
                source
            )
        elif isinstance(source, (Collection, dict)):
            from .prior import Collection
            instance = Collection._from_object(
                source
            )
        elif isinstance(source, str):
            from .instance import StringValue
            instance = StringValue._from_object(
                source
            )
        else:
            from .instance import Instance
            instance = Instance._from_object(
                source
            )
        instance.name = name
        return instance

    @property
    def _constructor_args(self):
        return set(
            inspect.getfullargspec(
                self.cls
            ).args[1:]
        )

    def _make_instance(self) -> object:
        """
        Create the real instance for this object
        """
        try:
            return object.__new__(self.cls)
        except TypeError as e:
            raise TypeError(
                f"Could not instantiate {self.name} of type {self.cls}"
            ) from e

    def __call__(self):
        """
        Create the real instance for this object, with child
        attributes attached.

        If the instance implements __setstate__ then this is
        called with a dictionary of instantiated children.
        """
        instance = self._make_instance()
        if hasattr(
                instance,
                "__setstate__"
        ):
            instance.__setstate__({
                child.name: child()
                for child in self.children
            })
        else:
            for child in self.children:
                setattr(
                    instance,
                    child.name,
                    child()
                )
        from autofit import ModelObject
        if isinstance(
                instance,
                ModelObject
        ) and (
                not hasattr(
                    instance, "id"
                ) or instance.id is None
        ):
            instance.id = instance.next_id()

        return instance

    def _add_children(
            self,
            items: Union[
                ItemsView[str, Any],
                Iterable[Tuple[str, Any]]
            ]
    ):
        """
        Add database representations of child attributes

        Parameters
        ----------
        items
            Attributes such as floats or priors that are associated
            with the real object
        """
        for key, value in items:
            if isinstance(
                    value,
                    property
            ) or key.startswith(
                "__"
            ) or key == "dtype":
                continue
            child = Object.from_object(
                value,
                name=key
            )
            self.children.append(
                child
            )

    class_path = sa.Column(
        sa.String
    )

    @property
    def cls(self) -> Type[object]:
        """
        The class of the real object
        """
        return get_class(
            self.class_path
        )

    @cls.setter
    def cls(self, cls: type):
        self.class_path = get_class_path(cls)



from scipy.interpolate import CubicSpline
from .abstract import AbstractInterpolator


class SplineInterpolator(AbstractInterpolator):
    """
    Interpolate data with a piecewise cubic polynomial which is twice continuously differentiable
    """

    @staticmethod
    def _interpolate(x, y, value):
        f = CubicSpline(x, y)
        return f(value)



from abc import ABC
from typing import Optional

from autofit.graphical.expectation_propagation import AbstractFactorOptimiser
from autofit.graphical.factor_graphs.factor import FactorKW
from autofit.mapper.prior_model.prior_model import AbstractPriorModel
from autofit.text.formatter import TextFormatter
from autofit.tools.namer import namer
from ..abstract import AbstractDeclarativeFactor


class AbstractModelFactor(FactorKW, AbstractDeclarativeFactor, ABC):
    @property
    def prior_model(self):
        return self._prior_model

    def __init__(
            self,
            prior_model: AbstractPriorModel,
            factor,
            optimiser: Optional[AbstractFactorOptimiser],
            prior_variable_dict,
            name=None,
            include_prior_factors=True
    ):
        """
        A factor in the graph that actually computes the likelihood of a model
        given values for each variable that model contains

        Parameters
        ----------
        prior_model
            A model with some dimensionality
        optimiser
            A custom optimiser that will be used to fit this factor specifically
            instead of the default optimiser
        """
        self._prior_model = prior_model
        self.optimiser = optimiser

        super().__init__(
            factor, **prior_variable_dict,
            name=name or namer(self.__class__.__name__),
        )
        self.include_prior_factors = include_prior_factors

    @property
    def info(self) -> str:
        """
        Info describing this factor. Same as model info with the factor name.

        Output as part of graph.info
        """
        return f"{self.name}\n\n{self.prior_model.info}"

    def make_results_text(self, model_approx) -> str:
        """
        Create a string describing the posterior values after this factor
        during or after an EPOptimisation.

        Parameters
        ----------
        model_approx: EPMeanField

        Returns
        -------
        A string containing the name of this factor with the names and
        values of each associated variable in the mean field.
        """
        arguments = {
            prior: model_approx.mean_field[prior] for prior in self.prior_model.priors
        }
        updated_model = self.prior_model.gaussian_prior_model_for_arguments(arguments)

        formatter = TextFormatter()

        for path, prior in updated_model.path_priors_tuples:
            formatter.add(path, prior.mean)
        return f"{self.name}\n\n{formatter.text}"



import logging
import os
from pathlib import Path

import matplotlib.pyplot as plt

from autofit.graphical.expectation_propagation.history import EPHistory

logger = logging.getLogger(__name__)


class Visualise:
    def __init__(self, ep_history: EPHistory, output_path: Path):
        """
        Handles visualisation of expectation propagation optimisation.

        This includes plotting key metrics such as Evidence and KL Divergence
        which are expected to converge.

        Parameters
        ----------
        ep_history
            A history describing previous optimisations by factor
        output_path
            The path that plots are written to
        """
        self.ep_history = ep_history
        self.output_path = output_path

        os.makedirs(output_path, exist_ok=True)

    def __call__(self):
        """
        Save a plot of Evidence and KL Divergence for the ep_history
        """
        fig, (evidence_plot, kl_plot) = plt.subplots(2)
        fig.suptitle("Evidence and KL Divergence")
        evidence_plot.plot(self.ep_history.evidences(), label="evidence")
        kl_plot.semilogy(self.ep_history.kl_divergences(), label="KL divergence")
        # for factor, factor_history in self.ep_history.items():
        #     evidence_plot.plot(
        #         factor_history.evidences, label=f"{factor.name} evidence"
        #     )
        #     kl_plot.plot(
        #         factor_history.kl_divergences, label=f"{factor.name} divergence"
        #     )
        evidence_plot.legend()
        kl_plot.legend()
        plt.savefig(str(self.output_path / "graph.png"))



import functools
import logging
from functools import wraps
from itertools import repeat
from typing import Tuple, Callable, List, Union, Optional

import numpy as np

from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.utils import Status
from autofit.non_linear.result import Result
from .ep_mean_field import EPMeanField
from ... import exc

logger = logging.getLogger(__name__)

EPCallBack = Callable[[Factor, EPMeanField, Status], bool]


def default_inf(func):
    """
    Decorator that catches HistoryException and returns inf.

    This used to give infinite divergence when there is insufficient
    history for a given factor.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except exc.HistoryException:
            return float("inf")

    return wrapper


def history_exception(func):
    @functools.wraps(func)
    def decorator(self):
        try:
            return func(self)
        except IndexError as e:
            raise exc.HistoryException(
                f"There have been no successful optimisations for factor {self.factor}"
            ) from e

    return decorator


class FactorHistory:
    def __init__(self, factor: Factor):
        """
        Tracks the history of a single factor

        Parameters
        ----------
        factor
            A factor in a graph undergoing optimisation
        """
        self.factor = factor
        self.history = list()

    def __call__(self, approx: EPMeanField, status: Status):
        """
        Add an optimisation result to the factor's history

        Parameters
        ----------
        approx
            The posterior mean field for an optimisation of the factor
        status
            Describes whether the optimisation was successful
        """
        self.history.append((approx, status))

    @property
    def successes(self) -> List[EPMeanField]:
        """
        A list of mean fields produced by successful optimisations
        """
        return [approx for approx, success in self.history if success]

    @property
    @history_exception
    def latest_result(self) -> Result:
        """
        The latest successful optimisation result for the factor.
        """
        return [status.result for _, status in self.history if status][0]

    @property
    @history_exception
    def latest_successful(self) -> EPMeanField:
        """
        A mean field for the last successful optimisation
        """
        return self.successes[-1]

    @property
    @history_exception
    def previous_successful(self) -> EPMeanField:
        """
        A mean field for the last-but-one successful optimisation
        """
        return self.successes[-2]

    @property
    def updates(self) -> List[EPMeanField]:
        """
        A list of mean fields produced by successful optimisations
        """
        return [approx for approx, status in self.history if status.updated]

    @property
    @history_exception
    def latest_update(self) -> EPMeanField:
        """
        Last updated mean field
        """
        return self.updates[-1]

    @property
    @history_exception
    def previous_update(self) -> EPMeanField:
        """
        Last-but-one updated mean field
        """
        return self.updates[-2]

    @default_inf
    def kl_divergence(self) -> Union[float, np.ndarray]:
        """
        The KL Divergence between the mean fields produced by the last
        two updates.

        If there are less than two successful optimisations then this is
        infinite.
        """
        return self.latest_update.mean_field.kl(self.previous_update.mean_field)

    @default_inf
    def evidence_divergence(self) -> Union[float, np.ndarray]:
        """
        The difference in the evidences between produced by the last two
        successful optimisations.

        If there are less than two successful optimisations then this is
        infinite.
        """
        return (
                self.latest_successful.log_evidence - self.previous_successful.log_evidence
        )

    @property
    def evidences(self) -> List[float]:
        """
        Evidences from successful optimisations with None as a placeholder
        when optimisation failed.
        """
        return [
            approx.log_evidence if status.updated else None
            for approx, status in self.history
        ]

    @property
    def kl_divergences(self) -> List[float]:
        """
        Evidences from successful optimisations with None as a placeholder
        when optimisation failed.
        """
        divergences = []
        for i in range(1, len(self.history)):
            previous, previous_success = self.history[i - 1]
            current, current_success = self.history[i]
            if previous_success.updated and current_success.updated:
                divergences.append(current.mean_field.kl(previous.mean_field))
            else:
                divergences.append(None)
        return divergences


class EPHistory:
    def __init__(
            self,
            callbacks: Tuple[EPCallBack, ...] = (),
            kl_tol: Optional[float] = 1e-1,
            evidence_tol: Optional[float] = None,
    ):
        """
        Track the history an an EP Optimization.

        Facilitates computation of termination metrics.

        Parameters
        ----------
        callbacks
            Custom callbacks which can be used to terminate
            optimisation
        kl_tol
            The minimum KL Divergence between the mean fields
            produced by two consecutive optimisations below which
            optimisation is terminated.

            If None this termination condition is ignored.
        evidence_tol
            The minimum difference in evidence between the mean
            fields produced by two consecutive optimisations below
            which optimisation is terminated.

            If None this termination condition is ignored.
        """
        self._callbacks = callbacks
        self.history = {}

        self.kl_tol = kl_tol
        self.evidence_tol = evidence_tol

    def __getitem__(self, factor: Factor) -> FactorHistory:
        """
        Retrieve the history associated with a given factor.
        """
        try:
            return self.history[factor]
        except KeyError:
            self.history[factor] = FactorHistory(factor)
            return self.history[factor]

    def items(self):
        return self.history.items()

    def __call__(
            self, factor: Factor, approx: EPMeanField, status: Status = Status()
    ) -> bool:
        """
        Add history for a given factor and determine whether optimisation
        should terminate.

        Parameters
        ----------
        factor
            A factor in the optimisation
        approx
            A mean field produced by optimisation of the factor
        status
            A status describing whether the optimisation was successful

        Returns
        -------
        A boolean indicating whether optimisation should terminate because
        divergence has dropped below a given tolerance or a callback evaluated
        to True.
        """
        self[factor](approx, status)
        if status.success:
            if any([callback(factor, approx, status) for callback in self._callbacks]):
                return True

            return self.is_converged(factor)

        return False

    def is_kl_converged(self, factor: Factor) -> bool:
        """
        True if the KL Divergence between the mean fields produced by
        two consecutive, successful optimisations is below the specified
        tolerance.
        """
        return self[factor].kl_divergence() < self.kl_tol

    def is_kl_evidence_converged(self, factor: Factor) -> bool:
        """
        True if the difference in evidence between produced by two consecutive,
        successful optimisations is below the specified tolerance.
        """
        evidence_divergence = self[factor].evidence_divergence()

        if evidence_divergence < 0:
            logger.warning(f"Evidence for factor {factor} has decreased")

        return abs(evidence_divergence) < self.evidence_tol

    def is_converged(self, factor: Factor) -> bool:
        """
        True if either convergence condition is met.
        """
        if self.kl_tol and self.is_kl_converged(factor):
            return True
        if self.evidence_tol and self.is_kl_evidence_converged(factor):
            return True
        return False

    def full_history(self, factor_order=None):
        factor_order = factor_order or self.history.keys()
        return [
            (f, approx, stat)
            for iteration in zip(
                *(zip(repeat(f), self.history[f].history) for f in factor_order)
            )
            for f, (approx, stat) in iteration
        ]

    def mean_field_history(self, factor_order=None):
        return [
            approx.mean_field for _, approx, _ in self.full_history(factor_order)
        ]

    def variable_history(self, factor_order=None):
        history = {}
        for mf in self.mean_field_history(factor_order):
            for v, m in mf.items():
                history.setdefault(v, []).append(m)

        return history

    def evidences(self, factor_order=None):
        history = self.full_history(factor_order)
        return [approx.log_evidence for _, approx, status in history]

    def kl_divergences(self, factor_order=None):
        mfs = self.mean_field_history(factor_order)
        n = len(self.history)
        return [m1.kl(m2) for m1, m2 in zip(mfs[n:], mfs)]



"""
Singleton to provide globally unique names

namer("hello") -> "hello0"
namer("hello") -> "hello1"
namer("world") -> "hello"
"""

from collections import Counter


class Namer:
    def __init__(self):
        self.counter = Counter()

    def __call__(self, name):
        number = self.counter[name]
        self.counter[name] += 1
        return f"{name}{number}"

    def reset(self):
        self.counter.clear()


namer = Namer()



import copy
import inspect
import json
import logging
import random
import types
from collections import defaultdict
from functools import wraps
from typing import Tuple, Optional, Dict, List, Iterable, Generator, Union, Type

import numpy as np

from autoconf import conf
from autoconf.exc import ConfigException
from autofit import exc
from autofit.mapper import model
from autofit.mapper.model import AbstractModel, frozen_cache
from autofit.mapper.prior import GaussianPrior
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior.deferred import DeferredArgument
from autofit.mapper.prior.tuple_prior import TuplePrior
from autofit.mapper.prior.width_modifier import WidthModifier
from autofit.mapper.prior_model.attribute_pair import DeferredNameValue
from autofit.mapper.prior_model.attribute_pair import (
    cast_collection,
    PriorNameValue,
    InstanceNameValue,
)
from autofit.mapper.prior_model.recursion import DynamicRecursionCache
from autofit.mapper.prior_model.util import PriorModelNameValue
from autofit.text import formatter as frm
from autofit.text.formatter import TextFormatter
from autofit.tools.util import info_whitespace
from autofit.tools.util import split_paths

logger = logging.getLogger(__name__)


class Limits:
    @staticmethod
    def for_class_and_attributes_name(cls, attribute_name):
        limit_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name, "gaussian_limits"]
        )
        return limit_dict["lower"], limit_dict["upper"]


class TuplePathModifier:
    def __init__(self, model_: "AbstractPriorModel"):
        """
        Modifies paths to priors contained in tuples.

        When a tuple is found in a signature a PriorTuple is created.

        The true path to a variable in the tuple is
        ("some", "preamble", "tuple_name", "tuple_name_0")

        Where 0 is the index of the tuple member. "tuple_name" must be
        removed from this path for some uses.

        When called, instances of this class remove the name of the tuple
        i.e:
        -> ("some", "preamble", "tuple_name_0")

        Parameters
        ----------
        model_
        """
        tuple_priors = model_.path_instance_tuples_for_class(TuplePrior)
        try:
            self.tuple_paths, _ = zip(*tuple_priors)
        except ValueError:
            self.tuple_paths = None

    def __call__(self, path):
        if self.tuple_paths is not None:
            if path[:-1] in self.tuple_paths:
                return path[:-2] + (path[-1],)
        return path


Path = Tuple[str, ...]


class MeanField:
    def __init__(self, prior_model: "AbstractPriorModel"):
        """
        Implements same interface as graphical code

        Parameters
        ----------
        prior_model
        """
        self.prior_model = prior_model

    def __getitem__(self, item):
        """
        Retrieve a prior by a prior with the same id
        """
        for prior in self.prior_model.priors:
            if prior == item:
                return prior
        raise KeyError(f"Could not find {item} in model")


def paths_to_tree(paths: List[Tuple[str, ...]], tree: Optional[dict] = None) -> dict:
    """
    Recursively convert a list of paths to a tree structure where common paths
    are matched.

    Parameters
    ----------
    paths
        A list of paths to attributes in the model.
    tree
        A tree already embedded in a parent tree.

    Returns
    -------
    A tree with depth max(map(len, paths))

    Examples
    --------
    paths_to_tree([
        ("one", "two", "three"),
        ("one", "two", "four"),
    ])

    gives

    {
        "one": {
            "two": {
                "three": {},
                "four": {}
            }
        }
    }
    """
    tree = tree or dict()
    for path in paths:
        if len(path) == 0:
            return tree
        first, *rest = path
        if first not in tree:
            child = dict()
            tree[first] = child
        tree[first] = paths_to_tree([rest], tree=tree[first])
    return tree


class AbstractPriorModel(AbstractModel):
    """
    Abstract model that maps a set of priors to a particular class. Must be
    overridden by any prior model so that the model mapper recognises its prior \
    model attributes.
    @DynamicAttrs
    """

    def __init__(self, label=None):
        super().__init__(label=label)
        self._assertions = list()

    def check_assertions(self, arguments: Dict[Prior, float]):
        """
        Check that all assertions are satisfied by the given arguments.

        Parameters
        ----------
        arguments
            A dictionary mapping priors to values

        Raises
        ------
        FitException
            If any assertion is not satisfied
        """
        failed_assertions = [
            assertion
            for assertion in self._assertions
            if assertion is False
            or assertion is not True
            and not assertion.instance_for_arguments(
                arguments,
            )
        ]
        number_of_failed_assertions = len(failed_assertions)
        if number_of_failed_assertions > 0:
            name_string = "\n".join(
                [
                    assertion.name
                    for assertion in failed_assertions
                    if hasattr(assertion, "name") and assertion.name is not None
                ]
            )
            raise exc.FitException(
                f"{number_of_failed_assertions} assertions failed!\n{name_string}"
            )

    def cast(
        self,
        value_dict: Dict["AbstractModel", dict],
        new_class: type,
    ) -> "AbstractPriorModel":
        """
        Cast models to a new type. Allows selected models in within this
        model to be given a new type and new arguments.

        Parameters
        ----------
        value_dict
            A dictionary mapping models to dictionaries of argument overrides
        new_class
            A new class to which specified models should be converted

        Returns
        -------
        A model where specified child models have been updated to a new class
        and new arguments
        """
        from .prior_model import Model

        updated = self

        for path, prior_model in self.path_instance_tuples_for_class(Model):
            try:
                model_value_dict = value_dict[prior_model]
                argument_dict = {
                    **dict(prior_model.direct_prior_tuples),
                    **dict(prior_model.direct_tuples_with_type(float)),
                    **model_value_dict,
                }
                updated = updated.replacing_for_path(
                    path, Model(new_class, **argument_dict)
                )

            except KeyError:
                pass

        return updated

    def without_attributes(self) -> "AbstractModel":
        """
        Returns a copy of this object with all priors, prior models and
        constants removed.
        """
        without_attributes = copy.copy(self)
        for key in self.__dict__:
            if not (key.startswith("_") or key in ("cls", "id")):
                delattr(without_attributes, key)
        return without_attributes

    def _with_paths(self, tree: dict) -> "AbstractModel":
        """
        Recursively generate a copy of this model retaining only objects
        specified by the tree.

        Parameters
        ----------
        tree
            A tree formed of dictionaries describing which components of the
            model should be retained.

        Returns
        -------
        A copy of this model with a subset of attributes
        """
        if len(tree) == 0:
            return self

        with_paths = self.without_attributes()
        for name, subtree in tree.items():
            # noinspection PyProtectedMember
            new_value = getattr(self, name)
            if isinstance(
                new_value,
                (
                    AbstractPriorModel,
                    TuplePrior,
                ),
            ):
                new_value = new_value._with_paths(subtree)
            setattr(with_paths, name, new_value)
        return with_paths

    @split_paths
    def with_paths(self, paths: List[Tuple[str, ...]]) -> "AbstractModel":
        """
        Recursively generate a copy of this model retaining only objects
        specified by the list of paths.

        Parameters
        ----------
        paths
            A list of tuples of strings each of which points to a retained attribute.
            All children of a given path are retained.

        Returns
        -------
        A copy of this model with a subset of attributes
        """
        return self._with_paths(paths_to_tree(paths))

    def _without_paths(self, tree: dict) -> "AbstractModel":
        """
        Recursively generate a copy of this model removing objects
        specified by the tree.

        Parameters
        ----------
        tree
            A tree formed of dictionaries describing which components of the
            model should be removed.

        Returns
        -------
        A copy of this model with a subset of attributes
        """
        without_paths = copy.deepcopy(self)
        for name, subtree in tree.items():
            # noinspection PyProtectedMember
            if len(subtree) == 0:
                delattr(without_paths, name)
            else:
                new_value = getattr(without_paths, name)
                if isinstance(
                    new_value,
                    (
                        AbstractPriorModel,
                        TuplePrior,
                    ),
                ):
                    new_value = new_value._without_paths(subtree)
                setattr(without_paths, name, new_value)
        return without_paths

    @split_paths
    def without_paths(self, paths: List[Tuple[str, ...]]) -> "AbstractModel":
        """
        Recursively generate a copy of this model retaining only objects
        not specified by the list of paths.

        Parameters
        ----------
        paths
            A list of tuples of strings each of which points to removed attribute.
            All children of a given path are removed.

        Returns
        -------
        A copy of this model with a subset of attributes
        """
        return self._without_paths(paths_to_tree(paths))

    def index(self, path: Tuple[str, ...]) -> int:
        """
        Retrieve the index of a given path in the model
        """
        return self.paths.index(path)

    @property
    def mean_field(self) -> MeanField:
        """
        Implements the same interface as the graphical code
        """
        return MeanField(self)

    @classmethod
    def from_json(cls, file: str):
        """
        Loads the model from a .json file, which was written using the model's dictionary (`dict`) attribute as
        follows:

        import json

        with open(filename, "w+") as f:
            json.dump(model.dict, f, indent=4)

        Parameters
        ----------
        file
            The path and name of the .json file the model is loaded from.

        Returns
        -------
        object
            The model, which may be a `Collection` of `Model` objects or a single `Model`.
        """

        with open(file) as json_file:
            model_dict = json.load(json_file)

        return cls.from_dict(d=model_dict)

    def add_assertion(self, assertion, name=""):
        """
        Assert that some relationship holds between physical values associated with
        priors at the point an instance is created. If this fails a FitException is
        raised causing the model to be re-sampled.
        Parameters
        ----------
        assertion
            An assertion that one prior must be greater than another.
        name
            A name describing the assertion that is logged when it is violated.
        """
        if assertion is True:
            return
        try:
            assertion.name = name
        except AttributeError:
            pass
        self._assertions.append(assertion)

    @property
    def name(self):
        return self.__class__.__name__

    # noinspection PyUnusedLocal
    @staticmethod
    def from_object(t, *args, **kwargs):
        if inspect.isclass(t):
            from .prior_model import Model

            obj = object.__new__(Model)
            obj.__init__(t, **kwargs)
        elif isinstance(t, list) or isinstance(t, dict):
            from autofit.mapper.prior_model import collection

            obj = object.__new__(collection.Collection)
            obj.__init__(t)
        else:
            obj = t
        return obj

    def take_attributes(self, source: object):
        """
        Take all attributes with a matching path from the source prior model.

        For example, if this prior model has a prior "one" and a matching prior
        is found associated with the source model then that attribute is attached
        to this model.

        If no matching attribute is found nothing happens.

        Parameters
        ----------
        source
            An instance or prior model from a previous search from which attributes
            are passed to this model.
        """

        def assert_no_assertions(obj):
            if len(getattr(obj, "_assertions", [])) > 0:
                raise AssertionError(
                    "take_attributes cannot be called once assertions have been added to the target"
                )

        assert_no_assertions(self)

        for path, _ in sum(
            map(self.path_instance_tuples_for_class, (Prior, float, TuplePrior)), []
        ):
            try:
                item = copy.copy(source)
                if isinstance(item, dict):
                    from autofit.mapper.prior_model.collection import Collection

                    item = Collection(item)
                for attribute in path:
                    if isinstance(attribute, int):
                        item = item[attribute]
                    else:
                        item = copy.copy(getattr(item, attribute))

                target = self
                for attribute in path[:-1]:
                    if isinstance(attribute, int):
                        target = target[attribute]
                    else:
                        target = getattr(target, attribute)
                    assert_no_assertions(target)

                attribute = path[-1]
                if isinstance(attribute, int):
                    target[attribute] = item
                else:
                    setattr(target, path[-1], item)
            except AttributeError:
                pass

    def instance_from_unit_vector(self, unit_vector, ignore_prior_limits=False):
        """
        Returns a ModelInstance, which has an attribute and class instance corresponding
        to every `Model` attributed to this instance.
        This method takes as input a unit vector of parameter values, converting each to
        physical values via their priors.
        Parameters
        ----------
        ignore_prior_limits
            If true then no exception is thrown if priors fall outside defined limits
        unit_vector: [float]
            A unit hypercube vector that is mapped to an instance of physical values via the priors.
        Returns
        -------
        model_instance : autofit.mapper.model.ModelInstance
            An object containing reconstructed model_mapper instances
        Raises
        ------
        exc.FitException
            If any assertion attached to this object returns False.
        """
        exception_tuples = self.attribute_tuples_with_type(ConfigException)
        if len(exception_tuples) > 0:
            for name, exception in exception_tuples:
                logger.exception(f"Could not load {name} because:\n\n{exception}")
            names = [name for name, _ in exception_tuples]
            raise ConfigException(
                f"No configuration was found for some attributes ({', '.join(names)})"
            )

        if self.prior_count != len(unit_vector):
            raise AssertionError(
                f"prior_count ({self.prior_count}) != len(unit_vector) {len(unit_vector)}"
            )

        arguments = dict(
            map(
                lambda prior_tuple, unit: (
                    prior_tuple.prior,
                    prior_tuple.prior.value_for(
                        unit,
                        ignore_prior_limits=ignore_prior_limits,
                    ),
                ),
                self.prior_tuples_ordered_by_id,
                unit_vector,
            )
        )

        return self.instance_for_arguments(
            arguments,
            ignore_assertions=ignore_prior_limits,
        )

    @property
    @cast_collection(PriorNameValue)
    @frozen_cache
    def unique_prior_tuples(self):
        """
        Returns
        -------
        prior_tuple_dict: [(Prior, PriorTuple)]
            The set of all priors associated with this mapper
        """
        return list(
            {
                prior_tuple[1]: prior_tuple
                for prior_tuple in self.attribute_tuples_with_type(Prior)
            }.values()
        )

    @property
    @cast_collection(PriorNameValue)
    @frozen_cache
    def prior_tuples_ordered_by_id(self):
        """
        Returns
        -------
        priors: [Prior]
            An ordered list of unique priors associated with this mapper
        """
        return sorted(
            list(self.unique_prior_tuples), key=lambda prior_tuple: prior_tuple.prior.id
        )

    @property
    def priors_ordered_by_id(self):
        return [prior for _, prior in self.prior_tuples_ordered_by_id]

    def vector_from_unit_vector(self, unit_vector, ignore_prior_limits=False):
        """
        Parameters
        ----------
        unit_vector: [float]
            A unit hypercube vector
        ignore_prior_limits
            Set to True to prevent an exception being raised if
            the physical value of a prior is outside the allowable
            limits

        Returns
        -------
        values: [float]
            A vector with values output by priors
        """
        return list(
            map(
                lambda prior_tuple, unit: prior_tuple.prior.value_for(
                    unit, ignore_prior_limits=ignore_prior_limits
                ),
                self.prior_tuples_ordered_by_id,
                unit_vector,
            )
        )

    def random_unit_vector_within_limits(
        self, lower_limit: float = 0.0, upper_limit: float = 1.0
    ) -> List[float]:
        """
        Generate a random vector of unit values by drawing uniform random values between 0 and 1.

        Returns
        -------
        unit_values
            A list of unit values constructed by taking random values from each prior.
        """
        return [
            random.uniform(
                max(lower_limit, prior.lower_unit_limit),
                min(upper_limit, prior.upper_unit_limit),
            )
            for prior in self.priors_ordered_by_id
        ]

    def random_vector_from_priors_within_limits(
        self, lower_limit: float = 0.0, upper_limit: float = 1.0
    ) -> List[float]:
        """
        Returns a random vector of physical values by drawing uniform random values between lower and upper limits
        defined as unit values, using the model priors to map them from unit values to physical values.

        This function guesses infinite unit vectors, until one is randomly sampled whose physical parameters are within
        the model's phyiscal limits.

        This is used for MCMC initialization, whereby the starting points of a walker(s) are confined to a restricted
        range of prior space. In particular, it is used for generate the "ball" initialization of Emcee.

        Parameters
        ----------
        lower_limit
            The lower limit as a unit value within which unit values (which are converted to physical values via the
            priors) are generated.
        upper_limit
            The upper limit as a unit value within which unit values (which are converted to physical values via the
            priors) are generated.

        Returns
        -------
        physical_values
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """
        vector = []

        for prior in self.priors_ordered_by_id:
            vector.append(
                prior.random(
                    lower_limit=lower_limit,
                    upper_limit=upper_limit,
                )
            )

        return vector

    def random_instance_from_priors_within_limits(
        self, lower_limit: float = 0.0, upper_limit: float = 1.0
    ):
        """
        Returns a random instance of physical values by drawing uniform random values between lower and upper limits
        defined as unit values, using the model priors to map them from unit values to physical values.

        This function guesses infinite unit vectors, until one is randomly sampled whose physical parameters are within
        the model's phyiscal limits. The final unit vector is then mapped to an instance.

        Parameters
        ----------
        lower_limit
            The lower limit as a unit value within which unit values (which are converted to physical values via the
            priors) are generated.
        upper_limit
            The upper limit as a unit value within which unit values (which are converted to physical values via the
            priors) are generated.

        Returns
        -------
        physical_values
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """

        vector = self.random_vector_from_priors_within_limits(
            lower_limit=lower_limit, upper_limit=upper_limit
        )

        return self.instance_from_vector(vector=vector)

    @property
    def random_vector_from_priors(self):
        """Generate a random vector of physical values by drawing uniform random values between 0 and 1 and using
        the model priors to map them from unit values to physical values.
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking random values from each prior.
        """
        return self.random_vector_from_priors_within_limits(
            lower_limit=0.0, upper_limit=1.0
        )

    @property
    def physical_values_from_prior_medians(self):
        """
        Returns
        -------
        physical_values: [float]
            A list of physical values constructed by taking the mean possible value from
            each prior.
        """
        return self.vector_from_unit_vector([0.5] * len(self.unique_prior_tuples))

    def instance_from_vector(self, vector, ignore_prior_limits=False):
        """
        Returns a ModelInstance, which has an attribute and class instance corresponding
        to every `Model` attributed to this instance.
        This method takes as input a physical vector of parameter values, thus omitting
        the use of priors.
        Parameters
        ----------
        vector: [float]
            A vector of physical parameter values that is mapped to an instance.
        ignore_prior_limits
            If True don't check that physical values are within expected limits.

        Returns
        -------
        model_instance : autofit.mapper.model.ModelInstance
            An object containing reconstructed model_mapper instances
        """
        if len(vector) != self.prior_count:
            raise AssertionError(
                f"Vector length {len(vector)} != prior count {self.prior_count}"
            )
        arguments = dict(
            map(
                lambda prior_tuple, physical_unit: (prior_tuple.prior, physical_unit),
                self.prior_tuples_ordered_by_id,
                vector,
            )
        )

        if not ignore_prior_limits:
            for prior, value in arguments.items():
                prior.assert_within_limits(value)

        return self.instance_for_arguments(
            arguments,
        )

    def has(self, cls: Union[Type, Tuple[Type, ...]]) -> bool:
        """
        Parameters
        ----------
        cls
            The type to check for

        Returns
        -------
        True iff this model contains an instance or model with the given
        type recursively. Includes models which have zero priors.
        """
        return self.has_instance(cls) or self.has_model(
            cls, include_zero_dimension=True
        )

    def has_instance(self, cls) -> bool:
        """
        True iff this model contains an instance of type
        cls, recursively.
        """
        return len(self.attribute_tuples_with_type(cls)) > 0

    def has_model(self, cls, include_zero_dimension=False) -> bool:
        """
        True iff this model contains a Model of type
        cls, recursively.
        """
        return (
            len(
                self.model_tuples_with_type(
                    cls,
                    include_zero_dimension=include_zero_dimension,
                )
            )
            > 0
        )

    def is_only_model(self, cls) -> bool:
        """
        True iff this model contains at least one Model
        of type cls and contains no PriorModels that are not
        of type cls, recursively.
        """
        from .prior_model import Model

        cls_models = self.model_tuples_with_type(cls)
        other_models = [
            value
            for _, value in self.attribute_tuples_with_type(Model)
            if value.prior_count > 0
        ]
        return len(cls_models) > 0 and len(cls_models) == len(other_models)

    def replacing(self, arguments):
        return self.mapper_from_partial_prior_arguments(arguments)

    @classmethod
    def product(cls, models: Iterable["AbstractPriorModel"]) -> "AbstractPriorModel":
        """
        Combine multiple models with the same structure by replacing priors with
        priors that contain a message which is the product of the messages of the
        priors with the same path in each model.

        Parameters
        ----------
        models
            A list of models to be combined

        Returns
        -------
        A model where each prior has a message which is the product of the messages
        associated with that prior across the models.
        """
        first, *rest = models

        arguments = dict()

        for path, prior in first.path_priors_tuples:
            for other in rest:
                prior = prior.with_message(
                    prior.message * other.object_for_path(path).message
                )
            arguments[prior] = prior
        return first.mapper_from_prior_arguments(arguments)

    def mapper_from_partial_prior_arguments(self, arguments):
        """
        Returns a new model mapper from a dictionary mapping existing priors to
        new priors, keeping existing priors where no mapping is provided.
        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping priors to priors
        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        original_prior_dict = {prior: prior for prior in self.priors}
        return self.mapper_from_prior_arguments({**original_prior_dict, **arguments})

    def mapper_from_prior_arguments(self, arguments):
        """
        Returns a new model mapper from a dictionary mapping existing priors to
        new priors.
        Parameters
        ----------
        arguments: {Prior: Prior}
            A dictionary mapping priors to priors
        Returns
        -------
        model_mapper: ModelMapper
            A new model mapper with updated priors.
        """
        logger.debug(f"Creating a new mapper from arguments")

        return self.gaussian_prior_model_for_arguments(arguments)

    def gaussian_prior_model_for_arguments(self, arguments):
        raise NotImplementedError()

    def mapper_from_gaussian_tuples(
        self, tuples, a=None, r=None, use_errors=True, use_widths=True, no_limits=False
    ):
        """
        The widths of the new priors are taken from the
        width_config. The new gaussian priors must be provided in the same order as
        the priors associated with model.
        If a is not None then all priors are created with an absolute width of a.
        If r is not None then all priors are created with a relative width of r.
        Parameters
        ----------
        no_limits
            If `True` generated priors have infinite limits
        r
            The relative width to be assigned to gaussian priors
        a
            print(tuples[i][1], width)
            The absolute width to be assigned to gaussian priors
        use_errors
            If True, the passed errors of the model components estimated in a previous `NonLinearSearch` (computed
            at the prior_passer.sigma value) are used to set the pass Gaussian Prior sigma value (if both width and
            passed errors are used, the maximum of these two values are used).
        use_widths
            If True, the minimum prior widths specified in the prior configs of the model components are used to
            set the passed Gaussian Prior sigma value (if both widths and passed errors are used, the maximum of
            these two values are used).
        tuples
            A list of tuples each containing the mean and width of a prior
        Returns
        -------
        mapper: ModelMapper
            A new model mapper with all priors replaced by gaussian priors.
        """

        prior_tuples = self.prior_tuples_ordered_by_id
        prior_class_dict = self.prior_class_dict
        arguments = {}

        for i, prior_tuple in enumerate(prior_tuples):
            prior = prior_tuple.prior
            cls = prior_class_dict[prior]
            mean, sigma = tuples[i]

            name = prior_tuple.name
            # Use the name of the collection for configuration when a prior's name
            # is just a number (i.e. its position in a collection)
            if name.isdigit():
                name = self.path_for_prior(prior_tuple.prior)[-2]

            width_modifier = (
                prior.width_modifier
                or WidthModifier.for_class_and_attribute_name(cls, name)
            )

            if a is not None and r is not None:
                raise exc.PriorException(
                    "Width of new priors cannot be both relative and absolute."
                )
            if a is not None:
                width = a
            elif r is not None:
                width = r * mean
            else:
                width = width_modifier(mean)

            if no_limits:
                limits = (float("-inf"), float("inf"))
            else:
                try:
                    limits = Limits.for_class_and_attributes_name(cls, name)
                except ConfigException:
                    limits = prior.limits

            if use_errors and not use_widths:
                sigma = tuples[i][1]
            elif not use_errors and use_widths:
                sigma = width
            elif use_errors and use_widths:
                sigma = max(tuples[i][1], width)
            else:
                raise exc.PriorException(
                    "use_passed_errors and use_widths are both False, meaning there is no "
                    "way to pass priors to set up the new model's Gaussian Priors."
                )

            new_prior = GaussianPrior(mean, sigma, *limits)
            new_prior.id = prior.id
            new_prior.width_modifier = prior.width_modifier
            arguments[prior] = new_prior

        return self.mapper_from_prior_arguments(arguments)

    def with_limits(self, limits: List[Tuple[float, float]]) -> "AbstractPriorModel":
        """
        Create a new instance of this model where each prior is updated to
        lie between new limits.

        Parameters
        ----------
        limits
            A list of pairs of lower and upper limits, one for each prior.

        Returns
        -------
        A new model with updated limits
        """
        return self.mapper_from_prior_arguments(
            {
                prior: prior.with_limits(*prior_limits)
                for prior, prior_limits in zip(self.priors_ordered_by_id, limits)
            }
        )

    def instance_from_prior_medians(self, ignore_prior_limits=False):
        """
        Returns a list of physical values from the median values of the priors.
        Returns
        -------
        physical_values : [float]
            A list of physical values
        """
        return self.instance_from_unit_vector(
            unit_vector=[0.5] * self.prior_count,
            ignore_prior_limits=ignore_prior_limits,
        )

    def log_prior_list_from_vector(
        self,
        vector: [float],
    ):
        """
        Compute the log priors of every parameter in a vector, using the Prior of every parameter.
        The log prior values are used by Emcee to map the log likelihood to the poserior of the model.
        Parameters
        ----------
        vector : [float]
            A vector of physical parameter values.
        Returns
        -------
        log_prior_list : []
            An list of the log prior value of every parameter.
        """
        return list(
            map(
                lambda prior_tuple, value: prior_tuple.prior.log_prior_from_value(
                    value=value
                ),
                self.prior_tuples_ordered_by_id,
                vector,
            )
        )

    def random_instance(self, ignore_prior_limits=False):
        """
        Returns a random instance of the model.
        """
        logger.debug(f"Creating a random instance")
        if ignore_prior_limits:
            return self.instance_from_unit_vector(
                unit_vector=[random.random() for _ in range(self.prior_count)],
                ignore_prior_limits=ignore_prior_limits,
            )
        return self.instance_for_arguments(
            {prior: prior.random() for prior in self.priors}
        )

    @staticmethod
    @DynamicRecursionCache()
    def from_instance(
        instance,
        model_classes: Union[type, Iterable[type]] = tuple(),
        exclude_classes: Union[type, Iterable[type]] = tuple(),
    ):
        """
        Recursively create a prior object model from an object model.

        Parameters
        ----------
        model_classes
            A tuple of classes that should be converted to a prior model
        exclude_classes
            A tuple of classes that should not be converted to a prior model
        instance
            A dictionary, list, class instance or model instance
        Returns
        -------
        abstract_prior_model
            A concrete child of an abstract prior model
        """
        from autofit.mapper.prior_model import collection

        if isinstance(instance, exclude_classes):
            return instance
        if isinstance(instance, (Prior, AbstractPriorModel)):
            return instance
        elif isinstance(instance, list):
            result = collection.Collection(
                [
                    AbstractPriorModel.from_instance(
                        item,
                        model_classes=model_classes,
                        exclude_classes=exclude_classes,
                    )
                    for item in instance
                ]
            )
        elif isinstance(instance, model.ModelInstance):
            from autofit.mapper import model_mapper

            result = model_mapper.ModelMapper()
            for key, value in instance.dict.items():
                setattr(
                    result,
                    key,
                    AbstractPriorModel.from_instance(
                        value,
                        model_classes=model_classes,
                        exclude_classes=exclude_classes,
                    ),
                )
        elif isinstance(instance, dict):
            result = collection.Collection(
                {
                    key: AbstractPriorModel.from_instance(
                        value,
                        model_classes=model_classes,
                        exclude_classes=exclude_classes,
                    )
                    for key, value in instance.items()
                }
            )
        elif isinstance(instance, (np.ndarray, types.FunctionType)):
            return instance
        else:
            from .prior_model import Model

            try:
                result = Model(
                    instance.__class__,
                    **{
                        key: AbstractPriorModel.from_instance(
                            value,
                            model_classes=model_classes,
                            exclude_classes=exclude_classes,
                        )
                        for key, value in instance.__dict__.items()
                        if key != "cls"
                    },
                )
            except AttributeError:
                return instance
        if isinstance(instance, model_classes):
            return result.as_model()
        return result

    def items(self):
        return (
            self.direct_prior_tuples
            + self.direct_instance_tuples
            + self.direct_prior_model_tuples
            + self.direct_tuple_priors
        )

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        return self.direct_tuples_with_type(Prior)

    @property
    @cast_collection(InstanceNameValue)
    def direct_instance_tuples(self):
        return self.direct_tuples_with_type(float)

    @property
    @cast_collection(PriorModelNameValue)
    def prior_model_tuples(self):
        return self.direct_tuples_with_type(AbstractPriorModel)

    @property
    @cast_collection(PriorModelNameValue)
    def direct_prior_model_tuples(self):
        return self.direct_tuples_with_type(AbstractPriorModel)

    @property
    @cast_collection(PriorModelNameValue)
    def direct_tuple_priors(self):
        return self.direct_tuples_with_type(TuplePrior)

    @property
    @cast_collection(PriorNameValue)
    def tuple_prior_tuples(self):
        """
        Returns
        -------
        tuple_prior_tuples: [(String, TuplePrior)]
        """
        return self.direct_tuples_with_type(TuplePrior)

    @property
    @cast_collection(PriorNameValue)
    def direct_prior_tuples(self):
        """
        Returns
        -------
        direct_priors: [(String, Prior)]
        """
        return self.direct_tuples_with_type(Prior)

    @property
    @cast_collection(DeferredNameValue)
    def direct_deferred_tuples(self):
        return self.direct_tuples_with_type(DeferredArgument)

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Prior))]
        """
        # noinspection PyUnresolvedReferences
        return self.attribute_tuples_with_type(Prior, ignore_children=True)

    def __eq__(self, other):
        return (
            isinstance(other, AbstractPriorModel)
            and self.direct_prior_model_tuples == other.direct_prior_model_tuples
        )

    @property
    @cast_collection(InstanceNameValue)
    def instance_tuples(self):
        """
        Returns
        -------
        instances: [(String, instance)]
        """
        return self.attribute_tuples_with_type(float, ignore_class=Prior)

    @property
    def prior_class_dict(self):
        from autofit.mapper.prior_model.annotation import AnnotationPriorModel

        d = {prior[1]: self.cls for prior in self.prior_tuples}
        for prior_model in self.prior_model_tuples:
            if not isinstance(prior_model[1], AnnotationPriorModel):
                d.update(prior_model[1].prior_class_dict)
        return d

    def _instance_for_arguments(
        self,
        arguments: Dict[Prior, float],
    ):
        raise NotImplementedError()

    def instance_for_arguments(
        self,
        arguments: Dict[Prior, float],
        ignore_assertions: bool = False,
    ):
        """
        Returns an instance of the model for a set of arguments

        Parameters
        ----------
        arguments
            Dictionary mapping priors to attribute analysis_path and value pairs
        ignore_assertions
            If True, assertions will not be checked

        Returns
        -------
            An instance of the class
        """
        if not (
            conf.instance["general"]["test"]["exception_override"] or ignore_assertions
        ):
            self.check_assertions(arguments)

        logger.debug(f"Creating an instance for arguments")
        return self._instance_for_arguments(
            arguments,
        )

    def path_for_name(self, name: str) -> Tuple[str, ...]:
        """
        Find the path to a prior in the model that matches
        a given name.

        For example, name_of_model_name_of_prior could match
        (name_of_model, name_of_prior). Unfortunately it is
        possible for ambiguity to occur. For example, another
        valid path could be (name, of_model, name_of_prior)
        in which case there is no way to determine which path
        actually matches the name.

        Parameters
        ----------
        name
            A name for a prior where names of models and the
            name of the prior have been joined by underscores.

        Returns
        -------
        A path to that model

        Raises
        ------
        AssertionError
            Iff no matching path is found
        """

        def _explode_path(path_):
            return tuple(string for part in path_ for string in part.split("_"))

        exploded = tuple(name.split("_"))
        for path, _ in self.path_priors_tuples:
            exploded_path = _explode_path(path)
            if exploded_path == exploded:
                return path

        for path, prior_tuple in self.path_instance_tuples_for_class(TuplePrior):
            for name, prior in prior_tuple.prior_tuples:
                total_path = path[:-1] + (name,)
                exploded_path = _explode_path(total_path)
                if exploded_path == exploded:
                    return path + (name,)
        raise AssertionError(f"No path was found matching {name}")

    def instance_from_prior_name_arguments(
        self, prior_name_arguments: Dict[str, float]
    ):
        """
        Instantiate the model from the names of priors and
        corresponding values.

        Parameters
        ----------
        prior_name_arguments
            The names of priors where names of models and the
            name of the prior have been joined by underscores,
            mapped to corresponding values.

        Returns
        -------
        An instance of the model
        """
        return self.instance_from_path_arguments(
            {
                self.path_for_name(name): value
                for name, value in prior_name_arguments.items()
            }
        )

    def instance_from_path_arguments(self, path_arguments: Dict[Tuple[str], float]):
        """
        Create an instance from a dictionary mapping paths to tuples
        to corresponding values.

        Parameters
        ----------
        path_arguments
            A dictionary mapping paths to priors to corresponding values.
            Note that, for linked priors, each path only needs to be
            specified once. If multiple paths for the same prior are
            specified then the value for the last path will be used.

        Returns
        -------
        An instance of the model
        """
        arguments = {
            self.object_for_path(path): value for path, value in path_arguments.items()
        }
        return self._instance_for_arguments(arguments)

    @property
    def prior_count(self) -> int:
        """
        How many unique priors does this model contain?
        """
        return len(self.unique_prior_tuples)

    @property
    def total_free_parameters(self) -> int:
        """
        Returns the prior count, but with a name that is more easy to interpret for users.
        """
        return self.prior_count

    @property
    def priors(self):
        return [prior_tuple.prior for prior_tuple in self.prior_tuples]

    @property
    def _prior_id_map(self):
        return {prior.id: prior for prior in self.priors}

    def prior_with_id(self, prior_id):
        return self._prior_id_map[prior_id]

    def name_for_prior(self, prior):
        for prior_model_name, prior_model in self.direct_prior_model_tuples:
            prior_name = prior_model.name_for_prior(prior)
            if prior_name is not None:
                return "{}_{}".format(prior_model_name, prior_name)
        prior_tuples = self.prior_tuples
        for name, p in prior_tuples:
            if p == prior:
                return name

    def __hash__(self):
        return self.id

    def __add__(self, other):
        if isinstance(other, Prior):
            return other + self

        result = copy.deepcopy(self)

        for key, value in other.__dict__.items():
            if not hasattr(result, key) or isinstance(value, Prior):
                setattr(result, key, value)
                continue
            result_value = getattr(result, key)
            if isinstance(value, AbstractPriorModel):
                if isinstance(result_value, AbstractPriorModel):
                    setattr(result, key, result_value + value)
                else:
                    setattr(result, key, value)

        return result

    def copy_with_fixed_priors(self, instance, excluded_classes=tuple()):
        """
        Recursively overwrite priors in the mapper with instance values from the
        instance except where the containing class is the descendant of a listed class.
        Parameters
        ----------
        excluded_classes
            Classes that should be left model
        instance
            The best fit from the previous search
        """
        mapper = copy.deepcopy(self)
        transfer_classes(instance, mapper, excluded_classes)
        return mapper

    @property
    def path_priors_tuples(self) -> List[Tuple[Path, Prior]]:
        path_priors_tuples = self.path_instance_tuples_for_class(Prior)
        return sorted(path_priors_tuples, key=lambda item: item[1].id)

    @property
    def paths(self) -> List[Path]:
        """
        A list of paths to all the priors in the model, ordered by their
        ids
        """
        return [path for path, _ in self.path_priors_tuples]

    def sort_priors_alphabetically(self, priors: Iterable[Prior]) -> List[Prior]:
        """
        Sort priors by their paths according to this model.

        Parameters
        ----------
        priors
            A set of priors

        Returns
        -------
        Those priors sorted alphabetically by path.
        """
        return sorted(priors, key=lambda prior: self.path_for_prior(prior))

    def path_for_prior(self, prior: Prior) -> Optional[Path]:
        """
        Find a path that points at the given tuple.
        Returns the first path or None if no path is found.
        Parameters
        ----------
        prior
            A prior representing what's known about some dimension of the model.
        Returns
        -------
        A path, a series of attributes that point to one location of the prior.
        """
        for path in self.all_paths_for_prior(prior):
            return path
        return None

    def all_paths_for_prior(self, prior: Prior) -> Generator[Path, None, None]:
        """
        Find all paths that points at the given tuple.

        Parameters
        ----------
        prior
            A prior representing what's known about some dimension of the model.
        Yields
        -------
        Paths, each a series of attributes that point to one location of the prior.
        """
        for path, path_prior in reversed(self.path_priors_tuples):
            if path_prior == prior:
                yield path

    @property
    def path_float_tuples(self):
        return self.path_instance_tuples_for_class(float, ignore_class=Prior)

    @property
    def unique_prior_paths(self):
        return [item[0] for item in self.unique_path_prior_tuples]

    @property
    def unique_path_prior_tuples(self):
        unique = {item[1]: item for item in self.path_priors_tuples}.values()
        return sorted(unique, key=lambda item: item[1].id)

    @property
    def prior_prior_model_dict(self):
        """
        Returns
        -------
        prior_prior_model_dict: {Prior: Model}
            A dictionary mapping priors to associated prior models. Each prior will only
            have one prior model; if a prior is shared by two prior models then one of
            those prior models will be in this dictionary.
        """
        return {
            prior: prior_model[1]
            for prior_model in self.prior_model_tuples + [("model", self)]
            for _, prior in prior_model[1].prior_tuples
        }

    def log_prior_list_from(self, parameter_lists: List[List]) -> List:
        return [
            sum(self.log_prior_list_from_vector(vector=vector))
            for vector in parameter_lists
        ]

    @property
    def info(self) -> str:
        """
        Use the priors that make up the model_mapper to generate information on each
        parameter of the overall model.
        This information is extracted from each priors *model_info* property.
        """

        formatter = TextFormatter(line_length=info_whitespace())

        for t in self.path_instance_tuples_for_class(
            (Prior, float, tuple), ignore_children=True
        ):
            formatter.add(*t)

        return "\n\n".join(
            [
                f"Total Free Parameters = {self.prior_count}",
                f"{self.parameterization}",
                formatter.text,
            ]
        )

    @property
    def order_no(self) -> str:
        """
        A string that can be used to order models by their
        parametrisation.

        Priors and constants are ordered by their paths and then
        joined into a string which means that models with higher
        associated values are consistently ordered later in a
        collection.
        """
        values = [
            str(float(value))
            for _, value in sorted(
                self.path_instance_tuples_for_class((Prior, float)), key=lambda t: t[0]
            )
        ]
        return ":".join(values)

    @property
    def parameterization(self) -> str:
        """
        Describes the path to each of the PriorModels, its class
        and its number of free parameters
        """
        from .prior_model import Model

        formatter = TextFormatter(line_length=info_whitespace())

        for t in self.path_instance_tuples_for_class(
            (
                Prior,
                float,
                tuple,
            ),
            ignore_children=True,
        ):
            for i in range(len(t[0])):
                path = t[0][:i]
                obj = self.object_for_path(path)
                if isinstance(obj, TuplePrior):
                    continue
                if isinstance(obj, AbstractPriorModel):
                    n = obj.prior_count
                else:
                    n = 0
                if isinstance(obj, Model):
                    name = obj.cls.__name__
                else:
                    name = type(obj).__name__

                formatter.add(("model",) + path, f"{name} (N={n})")

        return formatter.text

    @property
    def all_paths(self) -> List[Tuple[Path]]:
        """
        All possible paths to all priors grouped such that all
        paths to the same prior are collected together in a tuple.
        """
        if self.prior_count == 0:
            return []
        paths, _ = zip(*self.all_paths_prior_tuples)
        return paths

    @property
    def all_paths_prior_tuples(self) -> List[Tuple[Tuple[Path], Prior]]:
        """
        Maps a tuple containing all paths to a given prior to that prior.
        """
        prior_paths_dict = defaultdict(tuple)
        for path, prior in self.path_priors_tuples:
            prior_paths_dict[prior] += (path,)
        return sorted(
            [(paths, prior) for prior, paths in prior_paths_dict.items()],
            key=lambda item: item[1].id,
        )

    @property
    def all_names(self) -> List[Tuple[str]]:
        """
        All possible names for all priors grouped such that all
        names for a given prior are collected together in a tuple.
        """
        if self.prior_count == 0:
            return []
        names, _ = zip(*self.all_name_prior_tuples)
        return names

    @property
    def all_name_prior_tuples(self) -> List[Tuple[Tuple[str], Prior]]:
        """
        Maps a tuple containing all names for a given prior to that prior.
        """
        path_modifier = TuplePathModifier(self)
        return [
            (tuple("_".join(path_modifier(path)) for path in paths), prior)
            for paths, prior in self.all_paths_prior_tuples
        ]

    @property
    def model_component_and_parameter_names(self) -> List[str]:
        """The param_names vector is a list each parameter's analysis_path, and is used
        for *corner.py* visualization.
        The parameter names are determined from the class instance names of the
        model_mapper. Latex tags are properties of each model class."""
        prior_paths = self.unique_prior_paths

        tuple_filter = TuplePathModifier(self)

        prior_paths = list(map(tuple_filter, prior_paths))

        return ["_".join(path) for path in prior_paths]

    @property
    def parameter_names(self) -> List[str]:
        """
        The param_names vector is a list each parameter's analysis_path, and is used
        for *corner.py* visualization.

        The parameter names are determined from the class instance names of the
        model_mapper. Latex tags are properties of each model class.
        """
        return [parameter_name[-1] for parameter_name in self.unique_prior_paths]

    @property
    def parameter_labels(self) -> List[str]:
        """
        Returns a list of the label of every parameter in a model.

        This is used for displaying model results as text and for visualization with *corner.py*.

        The parameter labels are defined for every parameter of every model component in the config files label.ini and
        label_format.ini.
        """

        parameter_labels = []

        for parameter_name in self.parameter_names:
            parameter_label = frm.convert_name_to_label(
                parameter_name=parameter_name, name_to_label=True
            )
            parameter_labels.append(parameter_label)

        return parameter_labels

    @property
    def superscripts(self) -> List[str]:
        """
        Returns a list of the model component superscripts for every parameter in a model.


        The class superscript labels are defined as the name of every model component in the `ModelMapper`. For
        the example of a 1D Gaussian, if the model component name is `gaussian` three superscripts
        with this string (corresponding to the parameters `centre`, `normalization` and `sigma`) will
        be returned.

        For a `Collection`, the name of the inner model components are used.

        These superscripts may be overwritten by those returned from the `superscripts_config_overwrite` property,
        which optionally loads the superscripts from a `.json` config file. This allows high levels of customization
        in what superscripts are used.

        This is used for displaying model results as text and for visualization.
        """

        prior_paths = self.unique_prior_paths

        tuple_filter = TuplePathModifier(self)

        prior_paths = map(tuple_filter, prior_paths)

        superscripts = [path[-2] if len(path) > 1 else path[0] for path in prior_paths]

        return [
            superscript if not superscript_overwrite else superscript_overwrite
            for superscript, superscript_overwrite in zip(
                superscripts, self.superscripts_overwrite_via_config
            )
        ]

    @property
    def superscripts_overwrite_via_config(self) -> List[str]:
        """
        Returns a list of the model component superscripts for every parameter in a model, which can be used to
        overwrite the default superscripts used in the function above.

        The class superscript labels are defined for a model component in the config file `notation/label.ini`. By
        default, the model component names are used as superscripts (which are loaded via the method `superscripts`).
        These are overwritten by the superscripts loaded via a  config in this function. If no value is present in the
        config the model component names are used.

        For the example of a 1D Gaussian, when instatiated as a model component it is typically given the
        name `gaussian`. Thus, the string `gaussian` will be used as the supersript of every one its parameter labels
        (`centre`, `normalization` and `sigma`). However, if the config file `label.ini` reads `Gaussian=g`, every
        superscript for these parameters will instead be given the superscript `g`.

        This is used for displaying model results as text and for visualization with.
        """

        superscripts = []

        for prior_name, prior in self.prior_tuples_ordered_by_id:
            cls = self.prior_class_dict[prior]
            try:
                superscript = conf.instance["notation"]["label"]["superscript"][
                    cls.__name__
                ]

            except KeyError:
                superscript = ""

            superscripts.append(superscript)

        return superscripts

    @property
    def parameter_labels_with_superscripts(self) -> List[str]:
        """
        Returns a list of the latex parameter label and superscript of every parameter in a model.

        The parameter labels are defined for every parameter of every model component in the config file `label.ini`.
        This file can also be used to overwrite superscripts, that are assigned based on the model component name.

        This is used for displaying model results as text and for visualization, for example labelling parameters on a
        cornerplot.
        """

        return [
            f"{label}^{{\\rm {superscript}}}"
            for label, superscript in zip(self.parameter_labels, self.superscripts)
        ]

    @property
    def parameter_labels_with_superscripts_latex(self) -> List[str]:
        """
        Returns a list of the latex parameter label and superscript of every parameter in a model.

        The parameter labels are defined for every parameter of every model component in the config file `label.ini`.
        This file can also be used to overwrite superscripts, that are assigned based on the model component name.

        This is used for displaying model results as text and for visualization, for example labelling parameters on a
        cornerplot.
        """

        return [f"${label}$" for label in self.parameter_labels_with_superscripts]


def transfer_classes(instance, mapper, model_classes=None):
    """
    Recursively overwrite priors in the mapper with instance values from the
    instance except where the containing class is the descendant of a listed class.
    Parameters
    ----------
    model_classes
        Classes whose descendants should not be overwritten
    instance
        The best fit from the previous search
    mapper
        The prior model from the previous search
    """
    from autofit.mapper.prior_model.annotation import AnnotationPriorModel

    model_classes = model_classes or []
    for key, instance_value in instance.__dict__.items():
        try:
            mapper_value = getattr(mapper, key)
            if isinstance(mapper_value, Prior) or isinstance(
                mapper_value, AnnotationPriorModel
            ):
                setattr(mapper, key, instance_value)
                continue
            if not any(isinstance(instance_value, cls) for cls in model_classes):
                try:
                    transfer_classes(instance_value, mapper_value, model_classes)
                except AttributeError:
                    setattr(mapper, key, instance_value)
        except AttributeError:
            pass



import logging
import os
import pickle
from os import path
from pathlib import Path

import dill

from autofit.non_linear import abstract_search

original_create_file_handle = dill._dill._create_filehandle


def _create_file_handle(*args, **kwargs):
    """
    Handle FileNotFoundError when attempting to deserialize pickles
    using dill and return None instead.
    """
    try:
        return original_create_file_handle(*args, **kwargs)
    except pickle.UnpicklingError as e:
        if not isinstance(e.args[0], FileNotFoundError):
            raise e
        logging.warning(
            f"Could not create a handler for {e.args[0].filename} as it does not exist"
        )
        return None


dill._dill._create_filehandle = _create_file_handle


class Output:
    def __init__(self, directory: Path):
        self.directory = directory

    @property
    def pickle_path(self):
        return self.directory / "pickles"

    def __getattr__(self, item):
        """
        Attempt to load a pickle by the same name from the search output directory.

        dataset.pickle, meta_dataset.pickle etc.
        """
        try:
            with open(self.pickle_path / f"{item}.pickle", "rb") as f:
                return pickle.load(f)
        except FileNotFoundError:
            pass


class SearchOutput(Output):
    """
    @DynamicAttrs
    """

    def __init__(self, directory: str):
        """
        Represents the output of a single search. Comprises a metadata file and other dataset files.

        Parameters
        ----------
        directory
            The directory of the search
        """
        super().__init__(Path(directory))
        self.__search = None
        self.__model = None
        self.file_path = os.path.join(directory, "metadata")
        with open(self.file_path) as f:
            self.text = f.read()
            pairs = [line.split("=") for line in self.text.split("\n") if "=" in line]
            self.__dict__.update({pair[0]: pair[1] for pair in pairs})

    @property
    def child_analyses(self):
        """
        A list of child analyses loaded from the analyses directory
        """
        return list(map(Output, Path(self.directory).glob("analyses/*")))

    @property
    def model_results(self) -> str:
        """
        Reads the model.results file
        """
        with open(self.directory / "model.results") as f:
            return f.read()

    @property
    def mask(self):
        """
        A pickled mask object
        """
        with open(self.pickle_path / "mask.pickle", "rb") as f:
            return dill.load(f)

    @property
    def header(self) -> str:
        """
        A header created by joining the search name
        """
        phase = self.phase or ""
        dataset_name = self.dataset_name or ""
        return path.join(phase, dataset_name)

    @property
    def search(self) -> abstract_search.NonLinearSearch:
        """
        The search object that was used in this phase
        """
        if self.__search is None:
            try:
                with open(self.pickle_path / "search.pickle", "r+b") as f:
                    self.__search = pickle.loads(f.read())
            except (FileNotFoundError, ModuleNotFoundError) as e:
                print(self.pickle_path)
                logging.exception(e)
        return self.__search

    def child_values(self, name):
        """
        Get the values of a given key for all children
        """
        return [getattr(child, name) for child in self.child_analyses]

    @property
    def model(self):
        """
        The model that was used in this phase
        """
        if self.__model is None:
            try:
                with open(self.pickle_path / "model.pickle", "r+b") as f:
                    self.__model = pickle.loads(f.read())
            except (FileNotFoundError, ModuleNotFoundError) as e:
                print(self.pickle_path)
                logging.exception(e)
        return self.__model

    def __str__(self):
        return self.text

    def __repr__(self):
        return "<PhaseOutput {}>".format(self)



import math
from typing import Tuple, Union

import numpy as np
from scipy.special.cython_special import erfcinv
from scipy.stats import norm

from autoconf import cached_property
from autofit.mapper.operator import LinearOperator
from autofit.messages.abstract import AbstractMessage
from .composed_transform import TransformedMessage
from .transform import (
    phi_transform,
    log_transform,
    multinomial_logit_transform,
    log_10_transform,
)
from .. import exc


def is_nan(value):
    is_nan_ = np.isnan(value)
    if isinstance(is_nan_, np.ndarray):
        is_nan_ = is_nan_.all()
    return is_nan_


class NormalMessage(AbstractMessage):
    @cached_property
    def log_partition(self):
        eta1, eta2 = self.natural_parameters
        return -(eta1 ** 2) / 4 / eta2 - np.log(-2 * eta2) / 2

    log_base_measure = -0.5 * np.log(2 * np.pi)
    _support = ((-np.inf, np.inf),)
    _parameter_support = ((-np.inf, np.inf), (0, np.inf))

    def __init__(
        self,
        mean,
        sigma,
        lower_limit=-math.inf,
        upper_limit=math.inf,
        log_norm=0.0,
        id_=None,
    ):
        if (np.array(sigma) < 0).any():
            raise exc.MessageException("Sigma cannot be negative")

        super().__init__(
            mean,
            sigma,
            log_norm=log_norm,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            id_=id_,
        )
        self.mean, self.sigma = self.parameters

    def cdf(self, x):
        return norm.cdf(x, loc=self.mean, scale=self.sigma)

    def ppf(self, x):
        return norm.ppf(x, loc=self.mean, scale=self.sigma)

    @cached_property
    def natural_parameters(self):
        return self.calc_natural_parameters(self.mean, self.sigma)

    @staticmethod
    def calc_natural_parameters(mu, sigma):
        precision =  1 / sigma ** 2
        return np.array([mu * precision, -precision / 2])

    @staticmethod
    def invert_natural_parameters(natural_parameters):
        eta1, eta2 = natural_parameters
        mu = -0.5 * eta1 / eta2
        sigma = np.sqrt(-0.5 / eta2)
        return mu, sigma

    @staticmethod
    def to_canonical_form(x):
        return np.array([x, x ** 2])

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats):
        m1, m2 = suff_stats
        sigma = np.sqrt(m2 - m1 ** 2)
        return cls.calc_natural_parameters(m1, sigma)

    @cached_property
    def variance(self):
        return self.sigma ** 2

    def sample(self, n_samples=None):
        if n_samples:
            x = np.random.randn(n_samples, *self.shape)
            if self.shape:
                return x * self.sigma[None, ...] + self.mean[None, ...]
        else:
            x = np.random.randn(*self.shape)

        return x * self.sigma + self.mean

    def kl(self, dist):
        return (
            np.log(dist.sigma / self.sigma)
            + (self.sigma ** 2 + (self.mean - dist.mean) ** 2) / 2 / dist.sigma ** 2
            - 1 / 2
        )

    @classmethod
    def from_mode(
        cls, mode: np.ndarray, covariance: Union[float, LinearOperator] = 1.0, **kwargs
    ):
        if isinstance(covariance, LinearOperator):
            variance = covariance.diagonal()
        else:
            mode, variance = cls._get_mean_variance(mode, covariance)
        return cls(mode, np.abs(variance) ** 0.5, **kwargs)

    def _normal_gradient_hessian(
        self, x: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        # raise Exception
        shape = np.shape(x)
        if shape:
            x = np.asanyarray(x)
            deltax = x - self.mean
            hess_logl = -self.sigma ** -2
            grad_logl = deltax * hess_logl
            eta_t = 0.5 * grad_logl * deltax
            logl = self.log_base_measure + eta_t - np.log(self.sigma)

            if shape[1:] == self.shape:
                hess_logl = np.repeat(
                    np.reshape(hess_logl, (1,) + np.shape(hess_logl)), shape[0], 0
                )

        else:
            deltax = x - self.mean
            hess_logl = -self.sigma ** -2
            grad_logl = deltax * hess_logl
            eta_t = 0.5 * grad_logl * deltax
            logl = self.log_base_measure + eta_t - np.log(self.sigma)

        return logl, grad_logl, hess_logl

    def logpdf_gradient(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return self._normal_gradient_hessian(x)[:2]

    def logpdf_gradient_hessian(self, x: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        return self._normal_gradient_hessian(x)

    __name__ = "gaussian_prior"

    __default_fields__ = ("log_norm", "id_")

    def value_for(self, unit : float) -> float:
        """
        Returns a physical value from an input unit value according to the Gaussian distribution of the prior.

        Parameters
        ----------
        unit
            A unit value between 0 and 1.

        Returns
        -------
        value
            The unit value mapped to a physical value according to the prior.

        Examples
        --------

        prior = af.GaussianPrior(mean=1.0, sigma=2.0, lower_limit=0.0, upper_limit=2.0)

        physical_value = prior.value_for(unit=0.5)
        """
        return self.mean + (self.sigma * math.sqrt(2) * erfcinv(2.0 * (1.0 - unit)))

    def log_prior_from_value(self, value):
        """
        Returns the log prior of a physical value, so the log likelihood of a model evaluation can be converted to a
        posterior as log_prior + log_likelihood.
        This is used by certain non-linear searches (e.g. Emcee) in the log likelihood function evaluation.
        Parameters
        ----------
        value
            The physical value of this prior's corresponding parameter in a `NonLinearSearch` sample.
        """
        return (value - self.mean) ** 2.0 / (2 * self.sigma ** 2.0)

    def __str__(self):
        """
        The line of text describing this prior for the model_mapper.info file
        """
        return f"NormalMessage, mean = {self.mean}, sigma = {self.sigma}"

    def __repr__(self):
        return (
            "<NormalMessage id={} mean={} sigma={} "
            "lower_limit={} upper_limit={}>".format(
                self.id, self.mean, self.sigma, self.lower_limit, self.upper_limit
            )
        )
    
    @property
    def natural(self):
        return NaturalNormal.from_natural_parameters(
            self.natural_parameters * 0., **self._init_kwargs
        )

    def zeros_like(self) -> "AbstractMessage":
        return self.natural.zeros_like()


class NaturalNormal(NormalMessage):
    """
    Identical to the NormalMessage but allows non-normalised values,
    e.g negative or infinite variances
    """

    _parameter_support = ((-np.inf, np.inf), (-np.inf, 0))

    def __init__(
        self,
        eta1,
        eta2,
        lower_limit=-math.inf,
        upper_limit=math.inf,
        log_norm=0.0,
        id_=None,
    ):
        AbstractMessage.__init__(
            self,
            eta1,
            eta2,
            log_norm=log_norm,
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            id_=id_,
        )

    @cached_property
    def sigma(self):
        precision = -2 * self.parameters[1]
        return precision ** -0.5

    @cached_property
    def mean(self):
        return np.nan_to_num(-self.parameters[0] / self.parameters[1] / 2)

    @staticmethod
    def calc_natural_parameters(eta1, eta2):
        return np.array([eta1, eta2])

    @cached_property
    def natural_parameters(self):
        return self.calc_natural_parameters(*self.parameters)

    @classmethod
    def invert_sufficient_statistics(cls, suff_stats):
        m1, m2 = suff_stats
        precision = 1 / (m2 - m1 ** 2)
        return cls.calc_natural_parameters(m1 * precision, -precision / 2)

    @staticmethod
    def invert_natural_parameters(natural_parameters):
        return natural_parameters

    @classmethod
    def from_mode(
        cls, mode: np.ndarray, covariance: Union[float, LinearOperator] = 1.0, **kwargs
    ):
        if isinstance(covariance, LinearOperator):
            precision = covariance.inv().diagonal()
        else:
            mode, variance = cls._get_mean_variance(mode, covariance)
            precision = 1 / variance

        return cls(mode * precision, -precision / 2, **kwargs)

    zeros_like = AbstractMessage.zeros_like
    
    @property
    def natural(self):
        return self
    

UniformNormalMessage = TransformedMessage(NormalMessage(0, 1), phi_transform)

Log10UniformNormalMessage = TransformedMessage(UniformNormalMessage, log_10_transform)

LogNormalMessage = TransformedMessage(NormalMessage(0, 1), log_transform)

# Support is the simplex
MultiLogitNormalMessage = TransformedMessage(
    NormalMessage(0, 1), multinomial_logit_transform
)



from abc import ABC, abstractmethod
from typing import List, cast, Optional, Union

from autofit.graphical.declarative.factor.prior import PriorFactor
from autofit.graphical.expectation_propagation.ep_mean_field import EPMeanField
from autofit.graphical.factor_graphs.factor import Factor
from autofit.graphical.factor_graphs.graph import FactorGraph
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.variable import Variable
from autofit.text.formatter import TextFormatter


class DeclarativeGraphFormatter(ABC):
    def __init__(
            self,
            graph: "DeclarativeFactorGraph"
    ):
        self.graph = graph

    @abstractmethod
    def variable_formatter(self, variable):
        pass

    @property
    def info(self) -> str:
        """
        Describes the graph. Output in graph.info
        """
        prior_factor_info = "\n".join(
            map(
                self.info_for_prior_factor,
                self.graph.prior_factors
            )
        )
        analysis_factor_info = "\n\n".join(
            map(
                self.info_for_analysis_factor,
                self.graph.analysis_factors
            )
        )
        hierarchical_factor_info = "\n\n".join(
            map(
                self.info_for_hierarchical_factor,
                self.graph.hierarchical_factors
            )
        )
        string = f"""PriorFactors\n\n{prior_factor_info}\n\nAnalysisFactors\n\n{analysis_factor_info}"""
        if len(self.graph.hierarchical_factors) > 0:
            string = f"{string}\n\nHierarchicalFactors\n\n{hierarchical_factor_info}"
        return string

    def _related_factor_names(
            self,
            variable: Variable,
            excluded_factor: Optional[Factor] = None
    ) -> str:
        """
        Create a comma separated string describing factor names associated with
        the variable.

        Hierarchical factors are grouped.

        Parameters
        ----------
        variable
            A variable in the graph
        excluded_factor
            A factor which should not be included. e.g. the factor
            for which the variable is being checked.

        Returns
        -------
        A string describing the other factor's relationship to the variable.
        """
        from autofit.graphical.declarative.factor.hierarchical import _HierarchicalFactor

        related_factors = self.graph.related_factors(
            variable,
            excluded_factor=excluded_factor
        )

        names = set()

        for factor in related_factors:
            if isinstance(
                    factor,
                    _HierarchicalFactor
            ):
                names.add(
                    factor.distribution_model.name
                )
            else:
                names.add(
                    factor.name_for_variable(
                        variable
                    )
                )

        return ", ".join(sorted(names))

    def info_for_prior_factor(
            self,
            prior_factor: PriorFactor
    ) -> str:
        """
        A string describing a given PriorFactor in the context of this graph.
        """
        related_factor_names = self._related_factor_names(
            variable=prior_factor.variable,
            excluded_factor=prior_factor
        )

        formatter = TextFormatter()
        formatter.add(
            (f"{prior_factor.name} ({related_factor_names})",),
            self.variable_formatter(
                prior_factor.variable
            )
        )
        return formatter.text

    def info_for_analysis_factor(
            self,
            analysis_factor
    ) -> str:
        """
        A string describing a given AnalysisFactor in the context of this graph.
        """
        model = analysis_factor.prior_model
        formatter = TextFormatter()

        for path, prior in model.path_instance_tuples_for_class(
                Prior,
                ignore_children=True
        ):
            name = path[-1]
            related_factor_names = self._related_factor_names(
                prior,
                excluded_factor=analysis_factor
            )
            if len(related_factor_names) > 0:
                name = f"{name} ({related_factor_names})"
            path = path[:-1] + (name,)
            formatter.add(
                path,
                self.variable_formatter(
                    prior
                )
            )

        return f"{analysis_factor.name}\n\n{formatter.text}"

    def info_for_hierarchical_factor(
            self,
            hierarchical_factor
    ):
        distribution_model_info = self.info_for_analysis_factor(
            hierarchical_factor
        )

        formatter = TextFormatter()

        for factor in hierarchical_factor.factors:
            related_factor_names = self._related_factor_names(
                variable=factor.variable,
                excluded_factor=factor
            )
            formatter.add(
                (related_factor_names,),
                self.variable_formatter(
                    factor.variable
                )
            )

        return f"{distribution_model_info}\n\nDrawn Variables\n\n{formatter.text}"


class GraphInfoFormatter(DeclarativeGraphFormatter):
    """
    Formats a string that is output to graph.info
    """

    def variable_formatter(
            self,
            variable: Variable
    ):
        """
        Returns a string description of the variable
        """
        return str(variable)


class ResultsFormatter(DeclarativeGraphFormatter):
    def __init__(
            self,
            graph: "DeclarativeFactorGraph",
            model_approx: EPMeanField
    ):
        """
        Formats a string that is output to graph.results

        Parameters
        ----------
        graph
            A factor graph comprising AnalysisFactors
        model_approx
            An approximation of the model at the point of output
        """
        self.model_approx = model_approx
        super().__init__(graph)

    def variable_formatter(
            self,
            variable: Variable
    ):
        """
        Returns the mean value for the message corresponding to the
        variable
        """
        return self.model_approx.mean_field[
            variable
        ].mean


class DeclarativeFactorGraph(FactorGraph):
    @property
    def analysis_factors(self):
        """
        Analysis factors associated with this graph.
        """
        from .factor.analysis import AnalysisFactor
        return cast(
            List[AnalysisFactor],
            self._factors_with_type(
                AnalysisFactor
            )
        )

    @property
    def prior_factors(self) -> List[PriorFactor]:
        """
        Prior factors associated with this graph.
        """
        return cast(
            List[PriorFactor],
            self._factors_with_type(
                PriorFactor
            )
        )

    @property
    def hierarchical_factors(self):
        """
        Prior factors associated with this graph.
        """
        from autofit.graphical.declarative.factor.hierarchical import _HierarchicalFactor

        hierarchical_factor_set = set()

        for factor in self._factors_with_type(
                _HierarchicalFactor
        ):
            hierarchical_factor_set.add(
                cast(
                    _HierarchicalFactor,
                    factor
                ).distribution_model
            )

        name = lambda x: x.name
        return sorted(hierarchical_factor_set, key=name)

    @property
    def info(self) -> str:
        """
        Describes the graph. Output in graph.info
        """
        return GraphInfoFormatter(self).info

    def make_results_text(
            self,
            model_approx: Union[EPMeanField, AbstractPriorModel]
    ) -> str:
        """
        Generate text describing the graph w.r.t. a given model approximation
        """
        return ResultsFormatter(
            self,
            model_approx
        ).info



from itertools import chain, count
from typing import Optional, Tuple, Dict, Set, Union, List, TYPE_CHECKING
import operator
from abc import ABC, abstractmethod
from functools import wraps, reduce
from math import sqrt

import numpy as np
from autoconf import cached_property

from autofit.mapper.model_object import ModelObject

if TYPE_CHECKING:
    from autofit.mapper.operator import LinearOperator 
    from autofit.mapper.variable_operator import VariableFullOperator, VariableOperator 


class Plate:
    _ids = count()

    def __init__(self, name: Optional[str] = None):
        """
        Represents a dimension, such as number of observations, features or dimensions

        Parameters
        ----------
        name
            The name of this dimension
        """
        self.id = next(self._ids)
        self.name = name or f"plate_{self.id}"

    def __repr__(self):
        return f"{type(self).__name__}(name={self.name})"

    def __eq__(self, other):
        return isinstance(other, Plate) and self.id == other.id

    def __hash__(self):
        return self.id

    def __lt__(self, other):
        return self.id < other.id

    def __gt__(self, other):
        return self.id > other.id

    def make_index_seq(
        self,
        plates_index: Dict["Plate", Union[List[int], range, slice]],
        plate_sizes: Dict["Plate", int],
    ) -> Union[List[int], range]:
        seq = plates_index.get(self, range(plate_sizes[self]))
        if isinstance(seq, slice):
            seq = range(plate_sizes[self])[seq]

        return seq


def plates(*vals):
    """Helper function for making multiple plate objects

    Example
    -------
    x_, a_, b_, y_, z_ = plates("x, a, b", "y, z")
    """
    for val in vals:
        for v in val.split(","):
            yield Plate(v.strip())


class Variable(ModelObject):
    # __slots__ = ("name", "plates")

    def __init__(self, name: str = None, *plates: Plate, id_=None):
        """
        Represents a variable in the problem. This may be fixed data or some coefficient
        that we are optimising for.

        Parameters
        ----------
        name
            The name of this variable
        plates
            Representation of the dimensions of this variable
        """
        self.plates = plates
        super().__init__(id_=id_)
        self.name = name or f"{self.__class__.__name__.lower()}_{self.id}"

    def __repr__(self):
        args = ", ".join(chain([self.name], map(repr, self.plates)))
        return f"{self.__class__.__name__}({args})"

    def __hash__(self):
        return self.id

    def __len__(self):
        return len(self.plates)

    def __str__(self):
        return self.name

    @property
    def ndim(self) -> int:
        """
        How many dimensions does this variable have?
        """
        return len(self.plates)

    def make_indexes(
        self,
        plates_index: Dict["Plate", Union[List[int], range, slice]],
        plate_sizes: Dict["Plate", int],
    ) -> Tuple[np.ndarray, ...]:
        if any(p in plates_index for p in self.plates):
            return np.ix_(
                *(p.make_index_seq(plates_index, plate_sizes) for p in self.plates)
            )
        return ()


def variables(*vals):
    """Helper function for making multiple variable objects

    Example
    -------
    x_, a_, b_, y_, z_ = variables("x, a, b", "y, z")
    """
    for val in vals:
        for v in val.split(","):
            yield Variable(v.strip())


# This allows us to treat the class FactorValue as a variable
# that allows us to keep track of the FactorValue vs deterministic
# values when calculating gradients and jacobians
class VariableMetaClass(type, Variable):
    def __new__(cls, clsname, bases, attrs):
        newcls = super().__new__(cls, clsname, bases, attrs)
        Variable.__init__(newcls, clsname)
        return newcls


class FactorValue(np.ndarray, metaclass=VariableMetaClass):
    def __new__(cls, input_array, deterministic_values=None):
        obj = np.asarray(input_array).view(cls)
        obj.deterministic_values = deterministic_values or {}
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.deterministic_values = getattr(obj, "deterministic_values", None)

    @property
    def log_value(self) -> np.ndarray:
        if self.shape:
            return self.base
        else:
            return self.item()

    def __getitem__(self, index) -> np.ndarray:
        if index is type(self):
            return self
        elif isinstance(index, Variable):
            return self.deterministic_values[index]
        else:
            return super().__getitem__(index)

    def keys(self):
        return self.deterministic_values.keys()

    def values(self):
        return self.deterministic_values.values()

    def items(self):
        return self.deterministic_values.items()

    deterministic_variables = property(keys)

    def __repr__(self):
        r = np.ndarray.__repr__(self)
        return r[:-1] + ", " + repr(self.deterministic_values) + ")"

    def to_dict(self):
        return VariableData({FactorValue: self.base, **self.deterministic_values})


def broadcast_plates(
    value: np.ndarray,
    in_plates: Tuple[Plate, ...],
    out_plates: Tuple[Plate, ...],
    reducer: np.ufunc = np.sum,
) -> np.ndarray:
    """
    Extract the indices of a collection of plates then match
    the shape of the data to that shape.

    Parameters
    ----------
    value
        A value to broadcast
    in_plates
        Plates representing the dimensions of the values
    out_plates
        Plates representing the output dimensions
    reducer
        function to reduce excess plates over, default np.sum
        must take axis as keyword argument


    Returns
    -------
    The value reshaped to match the plates
    """
    n_in = len(in_plates)
    n_out = len(out_plates)
    shift = np.ndim(value) - n_in
    if not (0 <= shift <= 1):
        raise ValueError("dimensions of value incompatible with passed plates")

    in_axes = list(range(shift, n_in + shift))
    out_axes = []
    k = n_out + shift

    for plate in in_plates:
        try:
            out_axes.append(out_plates.index(plate) + shift)
        except ValueError:
            out_axes.append(k)
            k += 1

    moved_value = np.moveaxis(
        np.expand_dims(value, tuple(range(n_in + shift, k))),
        in_axes,
        out_axes,
    )
    return reducer(moved_value, axis=tuple(range(n_out + shift, k)))


def _get_variable_data_class(data):
    # So that these methods work with standard python dictionaries
    return type(data) if isinstance(data, VariableData) else VariableData


def _unary_op(op):
    @wraps(op)
    def __op__(self):
        cls = _get_variable_data_class(self)
        return cls({k: op(val) for k, val in self.items()})

    return __op__


def _binary_op(op, ravel=False):
    if ravel:

        @wraps(op)
        def __op__(self, other):
            cls = _get_variable_data_class(self)
            if isinstance(other, FactorValue):
                other = other.to_dict()

            if isinstance(other, dict):
                return cls(
                    {
                        k: op(np.ravel(self[k]), np.ravel(other[k]))
                        for k in self.keys() & other.keys()
                    }
                )
            elif isinstance(other, VariableLinearOperator):
                return op(dict(self), other)
            else:
                return cls({k: op(val, other) for k, val in self.items()})

    else:

        @wraps(op)
        def __op__(self, other):
            cls = _get_variable_data_class(self)
            if isinstance(other, FactorValue):
                other = other.to_dict()

            if isinstance(other, dict):
                return cls(
                    {k: op(self[k], other[k]) for k in self.keys() & other.keys()}
                )
            elif isinstance(other, VariableLinearOperator):
                return op(dict(self), other)
            else:
                return cls({k: op(val, other) for k, val in self.items()})

    return __op__


def rmul(x, y):
    return y * x


def rtruediv(x, y):
    return y / x


class VariableData(Dict[Variable, np.ndarray]):
    var_norm = _unary_op(np.linalg.norm)
    var_det = _unary_op(np.linalg.det)
    var_max = _unary_op(np.max)
    var_min = _unary_op(np.min)
    var_sum = _unary_op(np.sum)
    var_all = _unary_op(np.all)
    var_any = _unary_op(np.any)
    var_prod = _unary_op(np.prod)
    var_isfinite = _unary_op(np.isfinite)

    __abs__ = _unary_op(operator.abs)
    __pos__ = _unary_op(operator.pos)
    __neg__ = _unary_op(operator.neg)
    __lt__ = _binary_op(operator.lt)
    __le__ = _binary_op(operator.le)
    __eq__ = _binary_op(operator.eq)
    __ne__ = _binary_op(operator.ne)
    __gt__ = _binary_op(operator.gt)
    __ge__ = _binary_op(operator.ge)
    __and__ = _binary_op(operator.and_)
    __and__ = _binary_op(operator.or_)

    __add__ = _binary_op(operator.add)
    __radd__ = _binary_op(operator.add)
    __sub__ = _binary_op(operator.sub)
    __rsub__ = _binary_op(operator.sub)
    __mul__ = _binary_op(operator.mul)
    __rmul__ = _binary_op(rmul)
    __truediv__ = _binary_op(operator.truediv)
    __rtruediv__ = _binary_op(rtruediv)
    __pow__ = _binary_op(operator.pow)

    abs = __abs__
    neg = __neg__
    sub = __sub__
    add = __add__
    mul = __mul__
    div = __truediv__
    var_dot = _binary_op(np.dot, ravel=True)

    @property
    def T(self):
        return type(self)({k: val.T for k, val in self.items()})

    @property
    def shapes(self):
        return {k: np.shape(val) for k, val in self.items()}

    @property
    def sizes(self):
        return {k: np.size(val) for k, val in self.items()}

    @property
    def size(self):
        return sum(np.size(val) for val in self.values())

    def ravel(self):
        cls = _get_variable_data_class(self)
        return cls({k: np.ravel(val) for k, val in self.items()})

    def map(self, func, *args, **kwargs):
        cls = _get_variable_data_class(self)
        return cls(
            (k, func(val, *(arg[k] for arg in args), **kwargs))
            for k, val in self.items()
        )

    def reduce(self, func):
        return reduce(func, self.values())

    def mapreduce(self, func, op):
        return VariableData.map(self, func).reduce(op)

    def subset(self, variables):
        cls = _get_variable_data_class(self)
        return cls((v, self[v]) for v in variables if v in self)

    def sum(self) -> float:
        return sum(VariableData.var_sum(self).values())

    def prod(self) -> float:
        return VariableData.reduce(VariableData.var_prod(self), operator.mul)

    def all(self) -> bool:
        return all(VariableData.var_all(self).values())

    def any(self) -> bool:
        return any(VariableData.var_all(self).values())

    def det(self) -> float:
        return VariableData.var_det(self).reduce(operator.mul)

    def log_det(self) -> float:
        return VariableData.mapreduce(self, np.linalg.logdet, operator.add)

    def max(self) -> float:
        return max(VariableData.var_max(self).values())

    def min(self) -> float:
        return min(VariableData.var_min(self).values())

    def dot(self, other) -> float:
        return VariableData.var_dot(self, other).sum()

    def norm(self) -> float:
        return sqrt(VariableData.dot(self, self))

    def vecnorm(self, ord: Optional[float] = None) -> float:
        if ord:
            absval = VariableData.abs(self)
            if ord == np.Inf:
                return absval.max()
            elif ord == -np.Inf:
                return absval.min()
            else:
                return (absval ** ord).sum() ** (1.0 / ord)
        else:
            return VariableData.norm(self)

    def __repr__(self):
        name = type(self).__name__
        data_repr = dict.__repr__(self)
        return f"{name}({data_repr})"

    def merge(self, other):
        return VariableData({**self, **other})

    def plate_sizes(self):
        sizes = {}
        for v, val in self.items():
            shape = np.shape(val)
            assert len(shape) == len(v.plates), f"shape must match the number of plates of {v}"
            for p, s in zip(v.plates, shape):
                assert sizes.setdefault(p, s) == s, f"plate sizes must be consistent, {sizes[p]} != {s}"
        return sizes

    def full_like(self, fill_value, **kwargs):
        return type(self)({
            v: np.full_like(val, fill_value, **kwargs)
            for v, val in self.items()
        })

    def zeros_like(self, **kwargs):
        return self.full_like(0.)


class VariableLinearOperator(ABC):
    """Implements the functionality of a linear operator acting
    on a dictionary of values indexed by `Variable` objects
    """

    @abstractmethod
    def __mul__(self, x: VariableData) -> VariableData:
        pass

    @abstractmethod
    def __rtruediv__(self, x: VariableData) -> VariableData:
        pass

    @abstractmethod
    def __rmul__(self, x: VariableData) -> VariableData:
        pass

    @abstractmethod
    def ldiv(self, x: VariableData) -> VariableData:
        pass

    @property
    @abstractmethod
    def variables(self) -> VariableData:
        pass

    def __getitem__(self, variable) -> "LinearOperator":
        raise NotImplementedError()

    def get(self, variable, default=None):
        try:
            return self[variable]
        except KeyError:
            return default

    def __contains__(self, variable):
        return variable in self.variables

    def dot(self, x):
        return self * x

    __matmul__ = dot

    def inv(self) -> "InverseVariableOperator":
        return InverseVariableOperator(self)

    def quad(self, M: VariableData) -> VariableData:
        return (M * self).T * self

    def invquad(self, M: VariableData) -> VariableData:
        return (M / self).T / self

    @abstractmethod
    def update(self, *args: Tuple[VariableData, VariableData]):
        pass

    def lowrankupdate(self, *values: VariableData):
        return self.update(*((value, value) for value in values()))

    def lowrankdowndate(self, *values: VariableData):
        return self.update(*((value, VariableData.neg(value)) for value in values()))

    def blocks(self):
        return self.to_block().blocks()


class InverseVariableOperator(VariableLinearOperator):
    def __init__(self, op):
        self.operator = op

    def __mul__(self, x: VariableData) -> VariableData:
        return self.operator.ldiv(x)

    def __rtruediv__(self, x: VariableData) -> VariableData:
        return x * self.operator

    def __rmul__(self, x: VariableData) -> VariableData:
        return x / self.operator

    def ldiv(self, x: VariableData) -> VariableData:
        return self * x

    def quad(self, M: VariableData) -> VariableData:
        return self.operator.invquad(M)

    def invquad(self, M: VariableData) -> VariableData:
        return self.operator.quad(M)

    def inv(self) -> VariableLinearOperator:
        return self.operator

    @property
    def variables(self) -> Set[Variable]:
        return self.operator.variables

    @property
    def is_diagonal(self):
        return self.operator.is_diagonal

    @cached_property
    def log_det(self):
        return -self.operator.log_det

    def update(self, *args: Tuple[VariableData, VariableData]):
        # apply Sherman-Morrison formulat
        A = self.operator
        for (u, v) in args:
            A1u = A * u
            A1v = v * A
            vTA1u = -A1u.dot(v)
            A = A.update(A1u, A1v * vTA1u)

        return type(self)(A)

    def diagonalupdate(self, d: VariableData):
        A = self.operator.diagonalupdate(d ** -1)
        return type(self)(A)

    def to_full(self) -> "VariableFullOperator":
        full_op = self.operator.to_full()
        M = np.linalg.inv(full_op.operator.to_dense())
        return full_op.from_dense(M, full_op.param_shapes)

    def diagonal(self) -> VariableData:
        full_op = self.to_full()
        diag = full_op.operator.to_dense().diagonal()
        return full_op.param_shapes.unflatten(diag)

    def to_block(self) -> "VariableOperator":
        return self.to_full().to_block()

    def __getitem__(self, variable):
        return self.to_full()[variable]



import autofit as af

from autofit.mapper.mock.mock_model import MockClassx4
from autofit.non_linear.mock.mock_samples import MockSamples


def make_model_gaussian_x1():

    return af.Model(
        af.Gaussian
    )


def make_samples_x5():

    model = af.ModelMapper(mock_class_1=MockClassx4)

    parameters = [
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [0.0, 1.0, 2.0, 3.0],
        [21.0, 22.0, 23.0, 24.0],
        [0.0, 1.0, 2.0, 3.0],
    ]

    return MockSamples(
        model=model,
        sample_list=af.Sample.from_lists(
            model=model,
            parameter_lists=parameters,
            log_likelihood_list=[1.0, 2.0, 3.0, 10.0, 5.0],
            log_prior_list=[0.0, 0.0, 0.0, 0.0, 0.0],
            weight_list=[1.0, 1.0, 1.0, 1.0, 1.0],
        ),
    )


import numpy as np
import scipy
from typing import List, Dict

from autofit.non_linear.samples.pdf import SamplesPDF
from .abstract import AbstractInterpolator
from .query import Equality, InterpolatorPath
from autofit.non_linear.analysis.analysis import Analysis
from autofit.non_linear.nest.dynesty.static import DynestyStatic
from autofit.mapper.prior_model.prior_model import Model
from autofit.mapper.prior_model.collection import Collection
from autofit.mapper.prior.gaussian import GaussianPrior


class LinearRelationship:
    def __init__(self, m: float, c: float):
        """
        Describes a linear relationship between x and y, y = mx + c

        Parameters
        ----------
        m
            The gradient of the relationship
        c
            The y-intercept of the relationship
        """
        self.m = m
        self.c = c

    def __call__(self, x: float) -> float:
        """
        Calculate the value of y for a given value of x
        """
        return self.m * x + self.c

    def __str__(self):
        return f"y = {self.m}x + {self.c}"

    def __repr__(self):
        return str(self)


class CovarianceAnalysis(Analysis):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        inverse_covariance_matrix: np.ndarray,
    ):
        """
        An analysis class that describes a linear relationship between x and y, y = mx + c

        Parameters
        ----------
        x
            The x values (e.g. time)
        y
            The y values. This is a matrix comprising all the variables in the model at each x value
        inverse_covariance_matrix
        """
        self.x = x
        self.y = y
        self.inverse_covariance_matrix = inverse_covariance_matrix

    def _y(self, instance) -> np.ndarray:
        """
        Calculate the y values for a given instance of the model

        Parameters
        ----------
        instance
            An instance of the model

        Returns
        -------
        The y values
        """
        return np.array([relationship(x) for x in self.x for relationship in instance])

    def log_likelihood_function(self, instance) -> float:
        """
        Calculate the log likelihood of the model given an instance of the model

        Parameters
        ----------
        instance
            An instance of the model

        Returns
        -------
        The log likelihood
        """
        return -0.5 * (
            np.dot(
                self.y - self._y(instance),
                np.dot(self.inverse_covariance_matrix, self.y - self._y(instance)),
            )
        )


class CovarianceInterpolator(AbstractInterpolator):
    def __init__(
        self,
        samples_list: List[SamplesPDF],
    ):
        """
        An interpolator that uses the covariance matrix of a set of samples to find linear
        relationships between variables and some variable on which they depend (e.g. time)

        Parameters
        ----------
        samples_list
            A list of samples from which to calculate the covariance matrix
        """
        self.samples_list = samples_list
        # noinspection PyTypeChecker
        super().__init__([samples.max_log_likelihood() for samples in samples_list])

    def covariance_matrix(self) -> np.ndarray:
        """
        Calculate the covariance matrix of the samples

        This comprises covariance matrices for each sample, subsumed along the diagonal
        """
        matrices = [samples.covariance_matrix() for samples in self.samples_list]
        prior_count = self.samples_list[0].model.prior_count
        size = prior_count * len(self.samples_list)
        array = np.zeros((size, size))
        for i, matrix in enumerate(matrices):
            array[
                i * prior_count : (i + 1) * prior_count,
                i * prior_count : (i + 1) * prior_count,
            ] = matrix
        return array

    def inverse_covariance_matrix(self) -> np.ndarray:
        """
        Calculate the inverse covariance matrix of the samples
        """
        covariance_matrix = self.covariance_matrix()
        return scipy.linalg.inv(
            covariance_matrix + 1e-6 * np.eye(covariance_matrix.shape[0])
        )

    @staticmethod
    def _interpolate(x, y, value):
        raise NotImplementedError()

    def _analysis_for_value(self, value: Equality) -> CovarianceAnalysis:
        """
        Create a covariance analysis for a given value. That is an analysis that will
        optimise relationships between each variable and that value.

        Parameters
        ----------
        value
            The value to which the variables are related (e.g. time)

        Returns
        -------
        A linear analysis
        """
        x = []
        y = []
        for sample in sorted(
            self.samples_list,
            key=lambda s: value.path.get_value(s.max_log_likelihood()),
        ):
            # noinspection PyTypeChecker
            x.append(value.path.get_value(sample.max_log_likelihood()))
            y.extend([value for value in sample.max_log_likelihood(as_instance=False)])

        return CovarianceAnalysis(
            np.array(x),
            np.array(y),
            inverse_covariance_matrix=self.inverse_covariance_matrix(),
        )

    def _relationships_for_value(
        self,
        value: Equality,
        path_relationship_map=None,
    ) -> List[LinearRelationship]:
        """
        Calculate the linear relationships between each variable and a given value

        Parameters
        ----------
        value
            The value to which the variables are related (e.g. time)

        Returns
        -------
        A list of linear relationships
        """
        analysis = self._analysis_for_value(value)
        model = self.model(path_relationship_map=path_relationship_map or {})
        optimizer = DynestyStatic()
        result = optimizer.fit(model=model, analysis=analysis)
        return result.instance

    def __getitem__(self, value: Equality) -> float:
        return self.get(value)

    def get(
        self,
        value: Equality,
        path_relationship_map: Dict[InterpolatorPath, Model] = None,
    ) -> float:
        """
        Calculate the value of the variable for a given value of the variable to which it is related

        Parameters
        ----------
        path_relationship_map
            Specifies which model should be used for each relationship. Each parameter in the model
            is described by a LinearRelationship with GaussianPriors by default.
        value
            The value to which the variables are related (e.g. time)

        Returns
        -------
        The value of the variable at the given time
        """
        relationships = self._relationships_for_value(
            value,
            path_relationship_map=path_relationship_map,
        )
        model = self._single_model
        arguments = {
            prior: relationship(value.value)
            for prior, relationship in zip(
                model.priors_ordered_by_id,
                relationships,
            )
        }
        return model.instance_for_arguments(arguments)

    def _max_likelihood_samples_list(self) -> SamplesPDF:
        """
        Find the samples with the highest log likelihood
        """
        return max(self.samples_list, key=lambda s: max(s.log_likelihood_list))

    @property
    def _single_model(self):
        """
        The model from the point in the time series that gave the highest likelihood
        """
        return self._max_likelihood_samples_list().model

    def model(self, path_relationship_map=None) -> Collection:
        """
        Create a model that describes the linear relationships between each variable and the variable to which it is
        related
        """
        models = []
        single_model = self._single_model
        path_relationship_map = path_relationship_map or {}
        for prior in single_model.priors_ordered_by_id:
            path = single_model.path_for_prior(prior)
            if path in path_relationship_map:
                models.append(path_relationship_map[path])
                continue

            mean = prior.mean
            models.append(
                Model(
                    LinearRelationship,
                    m=GaussianPrior(mean=mean, sigma=mean),
                    c=GaussianPrior(mean=mean, sigma=mean),
                )
            )
        return Collection(models)



import logging

from .analysis import Analysis
from .combined import CombinedAnalysis, CombinedResult
from ..paths.abstract import AbstractPaths

from autofit.mapper.prior_model.collection import Collection

logger = logging.getLogger(__name__)


class IndexedAnalysis:
    def __init__(self, analysis: Analysis, index: int):
        """
        One instance in a collection corresponds to this analysis. That
        instance is identified by its index in the collection.

        Parameters
        ----------
        analysis
            An analysis that can be applied to an instance in a collection
        index
            The index of the instance that should be passed to the analysis
        """
        if isinstance(analysis, IndexedAnalysis):
            analysis = analysis.analysis
        self.analysis = analysis
        self.index = index

    def log_likelihood_function(self, instance):
        """
        Compute the log likelihood by taking the instance at the index
        """
        return self.analysis.log_likelihood_function(instance[self.index])

    # TODO : Add before fit methods here?

    def visualize(self, paths: AbstractPaths, instance, during_analysis):
        return self.analysis.visualize(paths, instance[self.index], during_analysis)

    def visualize_combined(self, analyses, paths: AbstractPaths, instance, during_analysis):
        return self.analysis.visualize_combined(analyses, paths, instance[self.index], during_analysis)

    def profile_log_likelihood_function(self, paths: AbstractPaths, instance):
        return self.profile_log_likelihood_function(paths, instance[self.index])

    def __getattr__(self, item):
        return getattr(self.analysis, item)

    def make_result(self, samples, model, sigma=3.0, use_errors=True, use_widths=True):
        return self.analysis.make_result(
            samples, model, sigma=sigma, use_errors=use_errors, use_widths=use_widths,
        )


class IndexCollectionAnalysis(CombinedAnalysis):
    def __init__(self, *analyses):
        """
        Collection of analyses where each analysis has a different
        corresponding model.

        Parameters
        ----------
        analyses
            A list of analyses each with a separate model
        """
        super().__init__(
            *[
                IndexedAnalysis(analysis, index,)
                for index, analysis in enumerate(analyses)
            ]
        )

    def make_result(self, samples, model, sigma=1.0, use_errors=True, use_widths=False):
        """
        Associate each model with an analysis when creating the result.
        """
        child_results = [
            analysis.make_result(
                samples.subsamples(model),
                model,
                sigma=sigma,
                use_errors=use_errors,
                use_widths=use_widths,
            )
            for model, analysis in zip(model, self.analyses)
        ]
        return CombinedResult(child_results)

    def modify_before_fit(self, paths: AbstractPaths, model: Collection):
        """
        Modify the analysis before fitting.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        model
            The model which is to be fitted.
        """
        return CombinedAnalysis(
            *(analysis.modify_before_fit(paths, model) for analysis in self.analyses)
        )

    def modify_after_fit(self, paths: AbstractPaths, model: Collection, result):
        """
        Modify the analysis after fitting.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        model
            The model which is to be fitted.
        result
            The result of the fit.
        """
        return CombinedAnalysis(
            *(
                analysis.modify_after_fit(paths, model, result)
                for analysis in self.analyses
            )
        )
